import hashlib
import logging
import shutil
import threading
from datetime import datetime
from pathlib import Path
from typing import Any, cast, Dict, Iterable, Optional, Sequence, Set, Tuple

from gi.repository import Gtk
from peewee import fn, prefetch

from sublime_music.adapters import api_objects as API

from . import models
from .. import (
    AlbumSearchQuery,
    CacheMissError,
    CachingAdapter,
    ConfigParamDescriptor,
    ConfigurationStore,
    ConfigureServerForm,
    SongCacheStatus,
    UIInfo,
)

KEYS = CachingAdapter.CachedDataKey


class FilesystemAdapter(CachingAdapter):
    """
    Defines an adapter which retrieves its data from the local filesystem.
    """

    # Configuration and Initialization Properties
    # ==================================================================================
    @staticmethod
    def get_ui_info() -> UIInfo:
        return UIInfo(
            name="Local Filesystem",
            description="Add a directory on your local filesystem",
            icon_basename="folder-music",
        )

    @staticmethod
    def get_configuration_form(config_store: ConfigurationStore) -> Gtk.Box:
        def verify_config_store() -> Dict[str, Optional[str]]:
            return {}

        return ConfigureServerForm(
            config_store,
            {
                "directory": ConfigParamDescriptor(
                    type=Path, description="Music Directory", pathtype="directory"
                )
            },
            verify_config_store,
        )

    @staticmethod
    def migrate_configuration(config_store: ConfigurationStore):
        pass

    def __init__(self, config: dict, data_directory: Path, is_cache: bool = False):
        self.data_directory = data_directory
        self.cover_art_dir = self.data_directory.joinpath("cover_art")
        self.music_dir = self.data_directory.joinpath("music")

        self.cover_art_dir.mkdir(parents=True, exist_ok=True)
        self.music_dir.mkdir(parents=True, exist_ok=True)

        self.is_cache = is_cache

        self.db_write_lock: threading.Lock = threading.Lock()
        database_filename = data_directory.joinpath("cache.db")
        models.database.init(database_filename)
        models.database.connect()

        with self.db_write_lock, models.database.atomic():
            models.database.create_tables(models.ALL_TABLES)
            self._migrate_db()

    def initial_sync(self):
        # TODO (#188) this is where scanning the fs should potentially happen?
        pass

    def shutdown(self):
        logging.info("Shutdown complete")

    # Database Migration
    # ==================================================================================
    def _migrate_db(self):
        pass

    # Usage and Availability Properties
    # ==================================================================================
    can_be_cached = False  # Can't be cached (there's no need).
    can_be_ground_truth = False  # TODO (#188)
    is_networked = False  # Doesn't access the network.

    # TODO (#200) make these dependent on cache state. Need to do this kinda efficiently
    can_get_cover_art_uri = True
    can_get_song_file_uri = True
    can_get_song_details = True
    can_get_artist = True
    can_get_albums = True
    can_get_album = True
    can_get_ignored_articles = True
    can_get_directory = True
    can_search = True

    def _can_get_key(self, cache_key: CachingAdapter.CachedDataKey) -> bool:
        if not self.is_cache:
            return True

        # As long as there's something in the cache (even if it's not valid) it may be
        # returned in a cache miss error.
        query = models.CacheInfo.select().where(models.CacheInfo.cache_key == cache_key)
        return query.count() > 0

    @property
    def can_get_playlists(self) -> bool:
        return self._can_get_key(KEYS.PLAYLISTS)

    @property
    def can_get_playlist_details(self) -> bool:
        return self._can_get_key(KEYS.PLAYLIST_DETAILS)

    @property
    def can_get_artists(self) -> bool:
        return self._can_get_key(KEYS.ARTISTS)

    @property
    def can_get_genres(self) -> bool:
        return self._can_get_key(KEYS.GENRES)

    supported_schemes = ("file",)
    supported_artist_query_types = {
        AlbumSearchQuery.Type.RANDOM,
        AlbumSearchQuery.Type.NEWEST,
        AlbumSearchQuery.Type.FREQUENT,
        AlbumSearchQuery.Type.RECENT,
        AlbumSearchQuery.Type.STARRED,
        AlbumSearchQuery.Type.ALPHABETICAL_BY_NAME,
        AlbumSearchQuery.Type.ALPHABETICAL_BY_ARTIST,
        AlbumSearchQuery.Type.YEAR_RANGE,
        AlbumSearchQuery.Type.GENRE,
    }

    # Data Helper Methods
    # ==================================================================================
    def _get_list(
        self,
        model: Any,
        cache_key: CachingAdapter.CachedDataKey,
        ignore_cache_miss: bool = False,
        where_clauses: Tuple[Any, ...] = None,
        order_by: Any = None,
    ) -> Sequence:
        result = model.select()
        if where_clauses is not None:
            result = result.where(*where_clauses)

        if order_by:
            result = result.order_by(order_by)

        if self.is_cache and not ignore_cache_miss:
            # Determine if the adapter has ingested data for this key before, and if
            # not, cache miss.
            if not models.CacheInfo.get_or_none(
                models.CacheInfo.valid == True,  # noqa: 712
                models.CacheInfo.cache_key == cache_key,
            ):
                raise CacheMissError(partial_data=result)
        return result

    def _get_object_details(
        self, model: Any, id: str, cache_key: CachingAdapter.CachedDataKey
    ) -> Any:
        obj = model.get_or_none(model.id == id)

        # Handle the case that this is the ground truth adapter.
        if not self.is_cache:
            if not obj:
                raise Exception(f"{model} with id={id} does not exist")
            return obj

        # If we haven't ingested data for this item before, or it's been invalidated,
        # raise a CacheMissError with the partial data.
        cache_info = models.CacheInfo.get_or_none(
            models.CacheInfo.cache_key == cache_key,
            models.CacheInfo.parameter == id,
            models.CacheInfo.valid == True,  # noqa: 712
        )
        if not cache_info:
            raise CacheMissError(partial_data=obj)

        return obj

    def _compute_song_filename(self, cache_info: models.CacheInfo) -> Path:
        try:
            if path_str := cache_info.path:
                # Make sure that the path is somewhere in the cache directory and a
                # malicious server (or MITM attacker) isn't trying to override files in
                # other parts of the system.
                path = self.music_dir.joinpath(path_str)
                if self.music_dir in path.parents:
                    return path
        except Exception:
            pass

        # Fall back to using the song file hash as the filename. This shouldn't happen
        # with good servers, but just to be safe.
        return self.music_dir.joinpath(cache_info.file_hash)

    # Data Retrieval Methods
    # ==================================================================================
    def get_cached_statuses(
        self, song_ids: Sequence[str]
    ) -> Dict[str, SongCacheStatus]:
        def compute_song_cache_status(song: models.Song) -> SongCacheStatus:
            try:
                file = song.file
                if self._compute_song_filename(file).exists():
                    if file.valid:
                        if file.cache_permanently:
                            return SongCacheStatus.PERMANENTLY_CACHED
                        return SongCacheStatus.CACHED

                    # The file is on disk, but marked as stale.
                    return SongCacheStatus.CACHED_STALE
            except Exception:
                pass

            return SongCacheStatus.NOT_CACHED

        cached_statuses = {song_id: SongCacheStatus.NOT_CACHED for song_id in song_ids}
        try:
            file_models = models.CacheInfo.select().where(
                models.CacheInfo.cache_key == KEYS.SONG_FILE
            )
            song_models = models.Song.select().where(models.Song.id.in_(song_ids))
            cached_statuses.update(
                {
                    s.id: compute_song_cache_status(s)
                    for s in prefetch(song_models, file_models)
                }
            )
        except Exception:
            pass

        return cached_statuses

    _playlists = None

    def get_playlists(self, ignore_cache_miss: bool = False) -> Sequence[API.Playlist]:
        if self._playlists is not None:
            return self._playlists

        self._playlists = self._get_list(
            models.Playlist,
            CachingAdapter.CachedDataKey.PLAYLISTS,
            ignore_cache_miss=ignore_cache_miss,
            order_by=fn.LOWER(models.Playlist.name),
        )
        return self._playlists

    def get_playlist_details(self, playlist_id: str) -> API.Playlist:
        return self._get_object_details(
            models.Playlist, playlist_id, CachingAdapter.CachedDataKey.PLAYLIST_DETAILS
        )

    def get_cover_art_uri(self, cover_art_id: str, scheme: str, size: int) -> str:
        cover_art = models.CacheInfo.get_or_none(
            models.CacheInfo.cache_key == CachingAdapter.CachedDataKey.COVER_ART_FILE,
            models.CacheInfo.parameter == cover_art_id,
        )
        if cover_art:
            filename = self.cover_art_dir.joinpath(str(cover_art.file_hash))
            if filename.exists():
                if cover_art.valid:
                    return str(filename)
                else:
                    raise CacheMissError(partial_data=str(filename))

        raise CacheMissError()

    def get_song_file_uri(self, song_id: str, schemes: Iterable[str]) -> str:
        song = models.Song.get_or_none(models.Song.id == song_id)
        if not song:
            if self.is_cache:
                raise CacheMissError()
            else:
                raise Exception(f"Song {song_id} does not exist.")

        try:
            if (song_file := song.file) and (
                filename := self._compute_song_filename(song_file)
            ):
                if filename.exists():
                    file_uri = f"file://{filename}"
                    if song_file.valid:
                        return file_uri
                    else:
                        raise CacheMissError(partial_data=file_uri)
        except models.CacheInfo.DoesNotExist:
            pass

        raise CacheMissError()

    def get_song_details(self, song_id: str) -> models.Song:
        return self._get_object_details(
            models.Song,
            song_id,
            CachingAdapter.CachedDataKey.SONG,
        )

    def get_artists(self, ignore_cache_miss: bool = False) -> Sequence[API.Artist]:
        return self._get_list(
            models.Artist,
            CachingAdapter.CachedDataKey.ARTISTS,
            ignore_cache_miss=ignore_cache_miss,
            where_clauses=(~(models.Artist.id.startswith("invalid:")),),
        )

    def get_artist(self, artist_id: str) -> API.Artist:
        return self._get_object_details(
            models.Artist, artist_id, CachingAdapter.CachedDataKey.ARTIST
        )

    def get_albums(
        self,
        query: AlbumSearchQuery,
        sort_direction: str = "ascending"
        # TODO (#208) deal with sort dir here?
    ) -> Sequence[API.Album]:
        strhash = query.strhash()
        query_result = models.AlbumQueryResult.get_or_none(
            models.AlbumQueryResult.query_hash == strhash
        )
        # If we've cached the query result, then just return it. If it's stale, then
        # return the old value as a cache miss error.
        if query_result and (
            cache_info := models.CacheInfo.get_or_none(
                models.CacheInfo.cache_key == CachingAdapter.CachedDataKey.ALBUMS,
                models.CacheInfo.parameter == strhash,
            )
        ):
            if cache_info.valid:
                return query_result.albums
            else:
                raise CacheMissError(partial_data=query_result.albums)

        # If we haven't ever cached the query result, try to construct one, and return
        # it as a CacheMissError result.

        sql_query = models.Album.select().where(
            ~(models.Album.id.startswith("invalid:"))
        )

        Type = AlbumSearchQuery.Type
        if query.type == Type.GENRE:
            assert query.genre
        genre_name = genre.name if (genre := query.genre) else None

        sql_query = {
            Type.RANDOM: sql_query.order_by(fn.Random()),
            Type.NEWEST: sql_query.order_by(models.Album.created.desc()),
            Type.FREQUENT: sql_query.order_by(models.Album.play_count.desc()),
            Type.STARRED: sql_query.where(models.Album.starred.is_null(False)).order_by(
                models.Album.name
            ),
            Type.ALPHABETICAL_BY_NAME: sql_query.order_by(models.Album.name),
            Type.ALPHABETICAL_BY_ARTIST: sql_query.order_by(models.Album.artist.name),
            Type.YEAR_RANGE: sql_query.where(
                models.Album.year.between(*query.year_range)
            ).order_by(models.Album.year, models.Album.name),
            Type.GENRE: sql_query.where(models.Album.genre == genre_name).order_by(
                models.Album.name
            ),
        }.get(query.type)

        raise CacheMissError(partial_data=sql_query)

    def get_all_albums(self) -> Sequence[API.Album]:
        return self._get_list(
            models.Album,
            CachingAdapter.CachedDataKey.ALBUMS,
            ignore_cache_miss=True,
            where_clauses=(
                ~(models.Album.id.startswith("invalid:")),
                models.Album.artist.is_null(False),
            ),
        )

    def get_album(self, album_id: str) -> API.Album:
        return self._get_object_details(
            models.Album, album_id, CachingAdapter.CachedDataKey.ALBUM
        )

    def get_ignored_articles(self) -> Set[str]:
        return set(
            map(
                lambda i: i.name,
                self._get_list(
                    models.IgnoredArticle, CachingAdapter.CachedDataKey.IGNORED_ARTICLES
                ),
            )
        )

    def get_directory(self, directory_id: str) -> models.Directory:
        return self._get_object_details(
            models.Directory, directory_id, CachingAdapter.CachedDataKey.DIRECTORY
        )

    def get_genres(self) -> Sequence[API.Genre]:
        return self._get_list(models.Genre, CachingAdapter.CachedDataKey.GENRES)

    def search(self, query: str) -> API.SearchResult:
        search_result = API.SearchResult(query)
        search_result.add_results("albums", self.get_all_albums())
        search_result.add_results("artists", self.get_artists(ignore_cache_miss=True))
        search_result.add_results(
            "songs",
            self._get_list(
                models.Song,
                CachingAdapter.CachedDataKey.SONG,
                ignore_cache_miss=True,
                where_clauses=(models.Song.artist.is_null(False),),
            ),
        )
        search_result.add_results(
            "playlists",
            self.get_playlists(ignore_cache_miss=True),
        )
        return search_result

    # Data Ingestion Methods
    # ==================================================================================
    def _strhash(self, string: str) -> str:
        return hashlib.sha1(bytes(string, "utf8")).hexdigest()

    def ingest_new_data(
        self,
        data_key: CachingAdapter.CachedDataKey,
        param: Optional[str],
        data: Any,
    ):
        assert self.is_cache, "FilesystemAdapter is not in cache mode!"

        # Wrap the actual ingestion function in a database lock, and an atomic
        # transaction.
        with self.db_write_lock, models.database.atomic():
            self._do_ingest_new_data(data_key, param, data)

    def invalidate_data(self, key: CachingAdapter.CachedDataKey, param: Optional[str]):
        assert self.is_cache, "FilesystemAdapter is not in cache mode!"

        # Wrap the actual ingestion function in a database lock, and an atomic
        # transaction.
        with self.db_write_lock, models.database.atomic():
            self._do_invalidate_data(key, param)

    def delete_data(self, key: CachingAdapter.CachedDataKey, param: Optional[str]):
        assert self.is_cache, "FilesystemAdapter is not in cache mode!"

        # Wrap the actual ingestion function in a database lock, and an atomic
        # transaction.
        with self.db_write_lock, models.database.atomic():
            self._do_delete_data(key, param)

    def _do_ingest_new_data(
        self,
        data_key: CachingAdapter.CachedDataKey,
        param: Optional[str],
        data: Any,
        partial: bool = False,
    ) -> Any:
        # TODO (#201): this entire function is not exactly efficient due to the nested
        # dependencies and everything. I'm not sure how to improve it, and I'm not sure
        # if it needs improving at this point.
        logging.debug(
            f"_do_ingest_new_data param={param} data_key={data_key} data={data}"
        )

        def getattrs(obj: Any, keys: Iterable[str]) -> Dict[str, Any]:
            return {k: getattr(obj, k) for k in keys}

        def setattrs(obj: Any, data: Dict[str, Any]):
            for k, v in data.items():
                if v is not None:
                    setattr(obj, k, v)

        def compute_file_hash(filename: str) -> str:
            file_hash = hashlib.sha1()
            with open(filename, "rb") as f:
                while chunk := f.read(8192):
                    file_hash.update(chunk)

            return file_hash.hexdigest()

        return_val = None

        # Set the cache info.
        now = datetime.now()
        cache_info, cache_info_created = models.CacheInfo.get_or_create(
            cache_key=(
                # In the case of SONG_FILE_PERMANENT, we have to use SONG_FILE as the
                # key in the database so everything matches up when querying.
                data_key
                if data_key != KEYS.SONG_FILE_PERMANENT
                else KEYS.SONG_FILE
            ),
            parameter=param,
            defaults={
                "cache_key": data_key,
                "parameter": param,
                "last_ingestion_time": now,
                # If it's partial data, then set it to be invalid so it will only be
                # used in the event that the ground truth adapter can't service the
                # request.
                "valid": not partial,
            },
        )
        if not cache_info_created:
            cache_info.valid = cache_info.valid or not partial
            cache_info.last_ingestion_time = now
            cache_info.save()

        if data_key == KEYS.ALBUM:
            album = cast(API.Album, data)
            album_id = album.id or f"invalid:{self._strhash(album.name)}"
            album_data = {
                "id": album_id,
                **getattrs(
                    album,
                    [
                        "name",
                        "created",
                        "duration",
                        "play_count",
                        "song_count",
                        "starred",
                        "year",
                    ],
                ),
                "genre": (
                    self._do_ingest_new_data(KEYS.GENRE, None, g)
                    if (g := album.genre)
                    else None
                ),
                "artist": (
                    self._do_ingest_new_data(KEYS.ARTIST, ar.id, ar, partial=True)
                    if (ar := album.artist)
                    else None
                ),
                "_songs": (
                    [
                        self._do_ingest_new_data(KEYS.SONG, s.id, s)
                        for s in album.songs or []
                    ]
                    if not partial
                    else None
                ),
                "_cover_art": (
                    self._do_ingest_new_data(
                        KEYS.COVER_ART_FILE, album.cover_art, data=None
                    )
                    if album.cover_art
                    else None
                ),
            }

            db_album, created = models.Album.get_or_create(
                id=album_id, defaults=album_data
            )

            if not created:
                setattrs(db_album, album_data)
                db_album.save()

            return_val = db_album

        elif data_key == KEYS.ALBUMS:
            albums = [
                self._do_ingest_new_data(KEYS.ALBUM, a.id, a, partial=True)
                for a in data
            ]
            album_query_result, created = models.AlbumQueryResult.get_or_create(
                query_hash=param, defaults={"query_hash": param, "albums": albums}
            )

            if not created:
                album_query_result.albums = albums
                try:
                    album_query_result.save()
                except ValueError:
                    # No save necessary.
                    pass

        elif data_key == KEYS.ARTIST:
            # Ingest similar artists.
            artist = cast(API.Artist, data)
            if artist.similar_artists:
                models.SimilarArtist.delete().where(
                    models.SimilarArtist.similar_artist.not_in(
                        [sa.id for sa in artist.similar_artists or []]
                    ),
                    models.Artist == artist.id,
                ).execute()
                models.SimilarArtist.insert_many(
                    [
                        {"artist": artist.id, "similar_artist": a.id, "order": i}
                        for i, a in enumerate(artist.similar_artists or [])
                    ]
                ).on_conflict_replace().execute()

            artist_id = artist.id or f"invalid:{self._strhash(artist.name)}"
            artist_data = {
                "id": artist_id,
                **getattrs(
                    artist,
                    [
                        "name",
                        "album_count",
                        "starred",
                        "biography",
                        "music_brainz_id",
                        "last_fm_url",
                    ],
                ),
                "albums": [
                    self._do_ingest_new_data(KEYS.ALBUM, a.id, a, partial=True)
                    for a in artist.albums or []
                ],
                "_artist_image_url": (
                    self._do_ingest_new_data(
                        KEYS.COVER_ART_FILE, artist.artist_image_url, data=None
                    )
                    if artist.artist_image_url
                    else None
                ),
            }

            db_artist, created = models.Artist.get_or_create(
                id=artist_id, defaults=artist_data
            )

            if not created:
                setattrs(db_artist, artist_data)
                db_artist.save()

            return_val = db_artist

        elif data_key == KEYS.ARTISTS:
            for a in data:
                self._do_ingest_new_data(KEYS.ARTIST, a.id, a, partial=True)
            models.Artist.delete().where(
                models.Artist.id.not_in([a.id for a in data])
                & ~models.Artist.id.startswith("invalid")
            ).execute()

        elif data_key == KEYS.COVER_ART_FILE:
            cache_info.file_id = param

            if data is not None:
                file_hash = compute_file_hash(data)
                cache_info.file_hash = file_hash

                # Copy the actual cover art file
                shutil.copy(str(data), str(self.cover_art_dir.joinpath(file_hash)))

        elif data_key == KEYS.DIRECTORY:
            api_directory = cast(API.Directory, data)
            directory_data: Dict[str, Any] = getattrs(
                api_directory, ["id", "name", "parent_id"]
            )

            if not partial:
                directory_data["directory_children"] = []
                directory_data["song_children"] = []
                for c in api_directory.children:
                    if hasattr(c, "children"):  # directory
                        directory_data["directory_children"].append(
                            self._do_ingest_new_data(
                                KEYS.DIRECTORY, c.id, c, partial=True
                            )
                        )
                    else:
                        directory_data["song_children"].append(
                            self._do_ingest_new_data(KEYS.SONG, c.id, c)
                        )

            directory, created = models.Directory.get_or_create(
                id=api_directory.id, defaults=directory_data
            )

            if not created:
                setattrs(directory, directory_data)
                directory.save()

            return_val = directory

        elif data_key == KEYS.GENRES:
            for g in data:
                self._do_ingest_new_data(KEYS.GENRE, None, g)

        elif data_key == KEYS.GENRE:
            api_genre = cast(API.Genre, data)
            genre_data = getattrs(api_genre, ["name", "song_count", "album_count"])
            genre, created = models.Genre.get_or_create(
                name=api_genre.name, defaults=genre_data
            )

            if not created:
                setattrs(genre, genre_data)
                genre.save()

            return_val = genre

        elif data_key == KEYS.IGNORED_ARTICLES:
            models.IgnoredArticle.insert_many(
                map(lambda s: {"name": s}, data)
            ).on_conflict_replace().execute()
            models.IgnoredArticle.delete().where(
                models.IgnoredArticle.name.not_in(data)
            ).execute()

        elif data_key == KEYS.PLAYLIST_DETAILS:
            api_playlist = cast(API.Playlist, data)
            playlist_data: Dict[str, Any] = {
                **getattrs(
                    api_playlist,
                    [
                        "id",
                        "name",
                        "song_count",
                        "duration",
                        "created",
                        "changed",
                        "comment",
                        "owner",
                        "public",
                    ],
                ),
                "_cover_art": (
                    self._do_ingest_new_data(
                        KEYS.COVER_ART_FILE, api_playlist.cover_art, None
                    )
                    if api_playlist.cover_art
                    else None
                ),
            }

            if not partial:
                # If it's partial, then don't ingest the songs.
                playlist_data["_songs"] = [
                    self._do_ingest_new_data(KEYS.SONG, s.id, s)
                    for s in api_playlist.songs
                ]

            playlist, playlist_created = models.Playlist.get_or_create(
                id=playlist_data["id"], defaults=playlist_data
            )

            # Update the values if the playlist already existed.
            if not playlist_created:
                setattrs(playlist, playlist_data)
                playlist.save()

            return_val = playlist

        elif data_key == KEYS.PLAYLISTS:
            self._playlists = None
            for p in data:
                self._do_ingest_new_data(KEYS.PLAYLIST_DETAILS, p.id, p, partial=True)
            models.Playlist.delete().where(
                models.Playlist.id.not_in([p.id for p in data])
            ).execute()

        elif data_key == KEYS.SEARCH_RESULTS:
            data = cast(API.SearchResult, data)
            for a in data._artists.values():
                self._do_ingest_new_data(KEYS.ARTIST, a.id, a, partial=True)

            for a in data._albums.values():
                self._do_ingest_new_data(KEYS.ALBUM, a.id, a, partial=True)

            for s in data._songs.values():
                self._do_ingest_new_data(KEYS.SONG, s.id, s, partial=True)

            for p in data._playlists.values():
                self._do_ingest_new_data(KEYS.PLAYLIST_DETAILS, p.id, p, partial=True)

        elif data_key == KEYS.SONG:
            api_song = cast(API.Song, data)
            song_data = getattrs(
                api_song, ["id", "title", "track", "year", "duration", "parent_id"]
            )
            song_data["genre"] = (
                self._do_ingest_new_data(KEYS.GENRE, None, g)
                if (g := api_song.genre)
                else None
            )
            song_data["artist"] = (
                self._do_ingest_new_data(KEYS.ARTIST, ar.id, ar, partial=True)
                if (ar := api_song.artist)
                else None
            )
            song_data["album"] = (
                self._do_ingest_new_data(KEYS.ALBUM, al.id, al, partial=True)
                if (al := api_song.album)
                else None
            )
            song_data["_cover_art"] = (
                self._do_ingest_new_data(
                    KEYS.COVER_ART_FILE,
                    api_song.cover_art,
                    data=None,
                )
                if api_song.cover_art
                else None
            )
            song_data["file"] = (
                self._do_ingest_new_data(
                    KEYS.SONG_FILE,
                    api_song.id,
                    data=(api_song.path, None, api_song.size),
                )
                if api_song.path
                else None
            )

            song, created = models.Song.get_or_create(
                id=song_data["id"], defaults=song_data
            )

            if not created:
                setattrs(song, song_data)
                song.save()

            return_val = song

        elif data_key == KEYS.SONG_FILE:
            cache_info.file_id = param

        elif data_key == KEYS.SONG_FILE_PERMANENT:
            cache_info.cache_permanently = True

        # Special handling for Song
        if data_key == KEYS.SONG_FILE and data:
            path, buffer_filename, size = data

            if path:
                cache_info.path = path

            if size:
                cache_info.size = size

            if buffer_filename:
                cache_info.file_hash = compute_file_hash(buffer_filename)

                # Copy the actual song file from the download buffer dir to the cache
                # dir.
                filename = self._compute_song_filename(cache_info)
                filename.parent.mkdir(parents=True, exist_ok=True)
                shutil.copy(str(buffer_filename), str(filename))

        cache_info.save()
        return return_val if return_val is not None else cache_info

    def _do_invalidate_data(
        self,
        data_key: CachingAdapter.CachedDataKey,
        param: Optional[str],
    ):
        logging.debug(f"_do_invalidate_data param={param} data_key={data_key}")
        models.CacheInfo.update({"valid": False}).where(
            models.CacheInfo.cache_key == data_key, models.CacheInfo.parameter == param
        ).execute()

        if data_key == KEYS.ALBUM:
            # Invalidate the corresponding cover art.
            if album := models.Album.get_or_none(models.Album.id == param):
                self._do_invalidate_data(KEYS.COVER_ART_FILE, album.cover_art)

        elif data_key == KEYS.ARTIST:
            # Invalidate the corresponding cover art and albums.
            if artist := models.Artist.get_or_none(models.Artist.id == param):
                self._do_invalidate_data(KEYS.COVER_ART_FILE, artist.artist_image_url)
                for album in artist.albums or []:
                    self._do_invalidate_data(
                        CachingAdapter.CachedDataKey.ALBUM, album.id
                    )

        elif data_key == KEYS.PLAYLIST_DETAILS:
            # Invalidate the corresponding cover art.
            if playlist := models.Playlist.get_or_none(models.Playlist.id == param):
                self._do_invalidate_data(KEYS.COVER_ART_FILE, playlist.cover_art)

        elif data_key == KEYS.SONG_FILE:
            # Invalidate the corresponding cover art.
            if song := models.Song.get_or_none(models.Song.id == param):
                self._do_invalidate_data(KEYS.COVER_ART_FILE, song.cover_art)

    def _do_delete_data(
        self, data_key: CachingAdapter.CachedDataKey, param: Optional[str]
    ):
        logging.debug(f"_do_delete_data param={param} data_key={data_key}")
        cache_info = models.CacheInfo.get_or_none(
            models.CacheInfo.cache_key == data_key,
            models.CacheInfo.parameter == param,
        )

        if data_key == KEYS.COVER_ART_FILE:
            if cache_info:
                self.cover_art_dir.joinpath(str(cache_info.file_hash)).unlink(
                    missing_ok=True
                )

        elif data_key == KEYS.PLAYLIST_DETAILS:
            # Delete the playlist and corresponding cover art.
            if playlist := models.Playlist.get_or_none(models.Playlist.id == param):
                if cover_art := playlist.cover_art:
                    self._do_delete_data(KEYS.COVER_ART_FILE, cover_art)

                playlist.delete_instance()

        elif data_key == KEYS.SONG_FILE:
            if cache_info:
                self._compute_song_filename(cache_info).unlink(missing_ok=True)

        elif data_key == KEYS.ALL_SONGS:
            shutil.rmtree(str(self.music_dir))
            shutil.rmtree(str(self.cover_art_dir))
            self.music_dir.mkdir(parents=True, exist_ok=True)
            self.cover_art_dir.mkdir(parents=True, exist_ok=True)

            models.CacheInfo.update({"valid": False}).where(
                models.CacheInfo.cache_key == KEYS.SONG_FILE
            ).execute()
            models.CacheInfo.update({"valid": False}).where(
                models.CacheInfo.cache_key == KEYS.COVER_ART_FILE
            ).execute()

        elif data_key == KEYS.EVERYTHING:
            self._do_delete_data(KEYS.ALL_SONGS, None)
            for table in models.ALL_TABLES:
                table.truncate_table()

        if cache_info:
            cache_info.valid = False
            cache_info.save()



from typing import List, Optional, Union

from peewee import (
    AutoField,
    BooleanField,
    ForeignKeyField,
    IntegerField,
    Model,
    prefetch,
    Query,
    SqliteDatabase,
    TextField,
)

from .sqlite_extensions import (
    CacheConstantsField,
    DurationField,
    SortedManyToManyField,
    TzDateTimeField,
)

database = SqliteDatabase(None)


# Models
# =============================================================================
class BaseModel(Model):
    class Meta:
        database = database


class CacheInfo(BaseModel):
    id = AutoField()
    valid = BooleanField(default=False)
    cache_key = CacheConstantsField()
    parameter = TextField(null=True, default="")
    # TODO (#2) actually use this for cache expiry.
    last_ingestion_time = TzDateTimeField(null=False)

    class Meta:
        indexes = ((("cache_key", "parameter"), True),)

    # Used for cached files.
    file_id = TextField(null=True)
    file_hash = TextField(null=True)
    size = IntegerField(null=True)
    path = TextField(null=True)
    cache_permanently = BooleanField(null=True)


class Genre(BaseModel):
    name = TextField(unique=True, primary_key=True)
    song_count = IntegerField(null=True)
    album_count = IntegerField(null=True)


class Artist(BaseModel):
    id = TextField(unique=True, primary_key=True)
    name = TextField(null=True)
    album_count = IntegerField(null=True)
    starred = TzDateTimeField(null=True)
    biography = TextField(null=True)
    music_brainz_id = TextField(null=True)
    last_fm_url = TextField(null=True)

    _artist_image_url = ForeignKeyField(CacheInfo, null=True)

    @property
    def artist_image_url(self) -> Optional[str]:
        try:
            return self._artist_image_url.file_id
        except Exception:
            return None

    @property
    def similar_artists(self) -> Query:
        return (
            Artist.select()
            .join(SimilarArtist, on=(SimilarArtist.similar_artist == Artist.id))
            .where(SimilarArtist.artist == self.id)
            .order_by(SimilarArtist.order)
        )


class SimilarArtist(BaseModel):
    artist = ForeignKeyField(Artist)
    similar_artist = ForeignKeyField(Artist)
    order = IntegerField()

    class Meta:
        # The whole thing is unique.
        indexes = ((("artist", "similar_artist", "order"), True),)


class Album(BaseModel):
    id = TextField(unique=True, primary_key=True)
    created = TzDateTimeField(null=True)
    duration = DurationField(null=True)
    name = TextField(null=True)
    play_count = IntegerField(null=True)
    song_count = IntegerField(null=True)
    starred = TzDateTimeField(null=True)
    year = IntegerField(null=True)

    artist = ForeignKeyField(Artist, null=True, backref="albums")
    genre = ForeignKeyField(Genre, null=True, backref="albums")

    _cover_art = ForeignKeyField(CacheInfo, null=True)

    @property
    def cover_art(self) -> Optional[str]:
        try:
            return self._cover_art.file_id
        except Exception:
            return None

    @property
    def songs(self) -> List["Song"]:
        albums = Album.select()
        artists = Album.select()
        return sorted(
            prefetch(self._songs, albums, artists),
            key=lambda s: (s.disc_number or 1, s.track),
        )


class AlbumQueryResult(BaseModel):
    query_hash = TextField(primary_key=True)
    albums = SortedManyToManyField(Album)


class IgnoredArticle(BaseModel):
    name = TextField(unique=True, primary_key=True)


class Directory(BaseModel):
    id = TextField(unique=True, primary_key=True)
    name = TextField(null=True)
    parent_id = TextField(null=True)

    _children: Optional[List[Union["Directory", "Song"]]] = None

    @property
    def children(self) -> List[Union["Directory", "Song"]]:
        if not self._children:
            self._children = list(
                Directory.select().where(Directory.parent_id == self.id)
            ) + list(Song.select().where(Song.parent_id == self.id))
        return self._children

    @children.setter
    def children(self, value: List[Union["Directory", "Song"]]):
        self._children = value


class Song(BaseModel):
    id = TextField(unique=True, primary_key=True)
    title = TextField()
    duration = DurationField(null=True)

    parent_id = TextField(null=True)
    album = ForeignKeyField(Album, null=True, backref="_songs")
    artist = ForeignKeyField(Artist, null=True)
    genre = ForeignKeyField(Genre, null=True, backref="songs")

    # figure out how to deal with different transcodings, etc.
    file = ForeignKeyField(CacheInfo, null=True)

    @property
    def size(self) -> Optional[int]:
        try:
            return self.file.size
        except Exception:
            return None

    @property
    def path(self) -> Optional[str]:
        try:
            return self.file.path
        except Exception:
            return None

    _cover_art = ForeignKeyField(CacheInfo, null=True)

    @property
    def cover_art(self) -> Optional[str]:
        try:
            return self._cover_art.file_id
        except Exception:
            return None

    track = IntegerField(null=True)
    disc_number = IntegerField(null=True)
    year = IntegerField(null=True)
    user_rating = IntegerField(null=True)
    starred = TzDateTimeField(null=True)


class Playlist(BaseModel):
    id = TextField(unique=True, primary_key=True)
    name = TextField()
    comment = TextField(null=True)
    owner = TextField(null=True)
    song_count = IntegerField(null=True)
    duration = DurationField(null=True)
    created = TzDateTimeField(null=True)
    changed = TzDateTimeField(null=True)
    public = BooleanField(null=True)

    _songs = SortedManyToManyField(Song, backref="playlists")

    @property
    def songs(self) -> List[Song]:
        albums = Album.select()
        artists = Album.select()
        return prefetch(self._songs, albums, artists)

    _cover_art = ForeignKeyField(CacheInfo, null=True)

    @property
    def cover_art(self) -> Optional[str]:
        try:
            return self._cover_art.file_id
        except Exception:
            return None


class Version(BaseModel):
    id = IntegerField(unique=True, primary_key=True)
    major = IntegerField()
    minor = IntegerField()
    patch = IntegerField()

    @staticmethod
    def is_less_than(semver: str) -> bool:
        major, minor, patch = map(int, semver.split("."))
        version, created = Version.get_or_create(
            id=0, defaults={"major": major, "minor": minor, "patch": patch}
        )
        if created:
            # There was no version before, definitely out-of-date
            return True

        return version.major < major or version.minor < minor or version.patch < patch

    @staticmethod
    def update_version(semver: str):
        major, minor, patch = map(int, semver.split("."))
        Version.update(major=major, minor=minor, patch=patch)


ALL_TABLES = (
    Album,
    AlbumQueryResult,
    AlbumQueryResult.albums.get_through_model(),
    Artist,
    CacheInfo,
    Directory,
    Genre,
    IgnoredArticle,
    Playlist,
    Playlist._songs.get_through_model(),
    SimilarArtist,
    Song,
    Version,
)



"""
These are the API objects that are returned by Subsonic.
"""

from dataclasses import dataclass, field
from datetime import datetime, timedelta
from typing import Any, Dict, List, Optional, Union

import dataclasses_json
from dataclasses_json import (
    config,
    dataclass_json,
    DataClassJsonMixin,
    LetterCase,
)
from dateutil import parser

from .. import api_objects as SublimeAPI

# Translation map
decoder_functions = {
    datetime: (lambda s: parser.parse(s) if s else None),
    timedelta: (lambda s: timedelta(seconds=s) if s else None),
}
encoder_functions = {
    datetime: (lambda d: datetime.strftime(d, "%Y-%m-%dT%H:%M:%S.%f%z") if d else None),
    timedelta: (lambda t: t.total_seconds() if t else None),
}

for type_, translation_function in decoder_functions.items():
    dataclasses_json.cfg.global_config.decoders[type_] = translation_function
    dataclasses_json.cfg.global_config.decoders[
        Optional[type_]  # type: ignore
    ] = translation_function

for type_, translation_function in encoder_functions.items():
    dataclasses_json.cfg.global_config.encoders[type_] = translation_function
    dataclasses_json.cfg.global_config.encoders[
        Optional[type_]  # type: ignore
    ] = translation_function


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Genre(SublimeAPI.Genre):
    name: str = field(metadata=config(field_name="value"))
    song_count: Optional[int] = None
    album_count: Optional[int] = None


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Album(SublimeAPI.Album):
    name: str
    id: Optional[str]
    cover_art: Optional[str] = None
    song_count: Optional[int] = None
    year: Optional[int] = None
    duration: Optional[timedelta] = None
    created: Optional[datetime] = None
    songs: List["Song"] = field(
        default_factory=list, metadata=config(field_name="song")
    )
    play_count: Optional[int] = None
    starred: Optional[datetime] = None

    # Artist
    artist: Optional["ArtistAndArtistInfo"] = field(init=False)
    _artist: Optional[str] = field(default=None, metadata=config(field_name="artist"))
    artist_id: Optional[str] = None

    # Genre
    genre: Optional[Genre] = field(init=False)
    _genre: Optional[str] = field(default=None, metadata=config(field_name="genre"))

    def __post_init__(self):
        # Initialize the cross-references
        self.artist = (
            None
            if not self.artist_id and not self._artist
            else ArtistAndArtistInfo(id=self.artist_id, name=self._artist)
        )
        self.genre = None if not self._genre else Genre(self._genre)


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class ArtistAndArtistInfo(SublimeAPI.Artist):
    name: str
    id: Optional[str]
    albums: List[Album] = field(
        default_factory=list, metadata=config(field_name="album")
    )
    album_count: Optional[int] = None
    cover_art: Optional[str] = None
    artist_image_url: Optional[str] = None
    starred: Optional[datetime] = None

    # Artist Info
    similar_artists: List["ArtistAndArtistInfo"] = field(default_factory=list)
    biography: Optional[str] = None
    music_brainz_id: Optional[str] = None
    last_fm_url: Optional[str] = None

    def __post_init__(self):
        if not self.album_count and len(self.albums) > 0:
            self.album_count = len(self.albums)
        if not self.artist_image_url:
            self.artist_image_url = self.cover_art

    def augment_with_artist_info(self, artist_info: Optional["ArtistInfo"]):
        if artist_info:
            self.similar_artists = artist_info.similar_artists
            self.biography = artist_info.biography
            self.last_fm_url = artist_info.last_fm_url
            self.artist_image_url = (
                artist_info.artist_image_url or self.artist_image_url
            )
            self.music_brainz_id = artist_info.music_brainz_id


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class ArtistInfo:
    similar_artists: List[ArtistAndArtistInfo] = field(
        default_factory=list, metadata=config(field_name="similarArtist")
    )
    biography: Optional[str] = None
    last_fm_url: Optional[str] = None
    artist_image_url: Optional[str] = field(
        default=None, metadata=config(field_name="largeImageUrl")
    )
    music_brainz_id: Optional[str] = None

    def __post_init__(self):
        if self.artist_image_url:
            placeholder_image_names = (
                "2a96cbd8b46e442fc41c2b86b821562f.png",
                "-No_image_available.svg.png",
            )
            for n in placeholder_image_names:
                if self.artist_image_url.endswith(n):
                    self.artist_image_url = ""
                    return


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Directory(SublimeAPI.Directory):
    id: str
    name: Optional[str] = None
    title: Optional[str] = None
    parent_id: Optional[str] = field(default=None, metadata=config(field_name="parent"))

    children: List[Union["Directory", "Song"]] = field(init=False)
    _children: List[Dict[str, Any]] = field(
        default_factory=list, metadata=config(field_name="child")
    )

    def __post_init__(self):
        if not isinstance(self.id, str):
            self.id = str(self.id)
        self.parent_id = (self.parent_id or "root") if self.id != "root" else None

        self.name = self.name or self.title
        self.children = [
            Directory.from_dict(c) if c.get("isDir") else Song.from_dict(c)
            for c in self._children
        ]


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Song(SublimeAPI.Song, DataClassJsonMixin):
    id: str
    title: str = field(metadata=config(field_name="name"))
    path: Optional[str] = None
    parent_id: Optional[str] = field(default=None, metadata=config(field_name="parent"))
    duration: Optional[timedelta] = None

    # Artist
    artist: Optional[ArtistAndArtistInfo] = field(init=False)
    _artist: Optional[str] = field(default=None, metadata=config(field_name="artist"))
    artist_id: Optional[str] = None

    # Album
    album: Optional[Album] = field(init=False)
    _album: Optional[str] = field(default=None, metadata=config(field_name="album"))
    album_id: Optional[str] = None

    # Genre
    genre: Optional[Genre] = field(init=False)
    _genre: Optional[str] = field(default=None, metadata=config(field_name="genre"))

    track: Optional[int] = None
    disc_number: Optional[int] = None
    year: Optional[int] = None
    size: Optional[int] = None
    cover_art: Optional[str] = None
    user_rating: Optional[int] = None
    starred: Optional[datetime] = None

    def __post_init__(self):
        if not isinstance(self.id, str):
            self.id = str(self.id)
        self.parent_id = (self.parent_id or "root") if self.id != "root" else None
        self.artist = (
            None
            if not self._artist
            else ArtistAndArtistInfo(id=self.artist_id, name=self._artist)
        )
        self.album = (
            None if not self._album else Album(id=self.album_id, name=self._album)
        )
        self.genre = None if not self._genre else Genre(self._genre)


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Playlist(SublimeAPI.Playlist):
    id: str
    name: str
    songs: List[Song] = field(default_factory=list, metadata=config(field_name="entry"))
    song_count: Optional[int] = field(default=None)
    duration: Optional[timedelta] = field(default=None)
    created: Optional[datetime] = None
    changed: Optional[datetime] = None
    comment: Optional[str] = None
    owner: Optional[str] = None
    public: Optional[bool] = None
    cover_art: Optional[str] = None

    def __post_init__(self):
        if not isinstance(self.id, str):
            self.id = str(self.id)
        if self.songs is None:
            return
        if self.song_count is None:
            self.song_count = len(self.songs)

        if self.duration is None:
            self.duration = timedelta(
                seconds=sum(
                    s.duration.total_seconds() if s.duration else 0 for s in self.songs
                )
            )


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class PlayQueue(SublimeAPI.PlayQueue):
    songs: List[Song] = field(default_factory=list, metadata=config(field_name="entry"))
    position: timedelta = timedelta(0)
    username: Optional[str] = None
    changed: Optional[datetime] = None
    changed_by: Optional[str] = None
    value: Optional[str] = None
    current: Optional[str] = None
    current_index: Optional[int] = None

    def __post_init__(self):
        if pos := self.position:
            # The position for this endpoint is in milliseconds instead of seconds
            # because the Subsonic API is sometime stupid.
            self.position = pos / 1000
        if cur := self.current:
            self.current_index = [s.id for s in self.songs].index(str(cur))


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Index:
    name: str
    artist: List[Dict[str, Any]] = field(default_factory=list)


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class IndexID3:
    name: str
    artist: List[ArtistAndArtistInfo] = field(default_factory=list)


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class ArtistsID3:
    ignored_articles: str
    index: List[IndexID3] = field(default_factory=list)


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class AlbumList2:
    album: List[Album] = field(default_factory=list)


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Genres:
    genre: List[Genre] = field(default_factory=list)


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Indexes:
    ignored_articles: str
    index: List[Index] = field(default_factory=list)


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class Playlists:
    playlist: List[Playlist] = field(default_factory=list)


@dataclass_json(letter_case=LetterCase.CAMEL)
@dataclass
class SearchResult3:
    artist: List[ArtistAndArtistInfo] = field(default_factory=list)
    album: List[Album] = field(default_factory=list)
    song: List[Song] = field(default_factory=list)


@dataclass
class Response(DataClassJsonMixin):
    """The base Subsonic response object."""

    artists: Optional[ArtistsID3] = None
    artist: Optional[ArtistAndArtistInfo] = None
    artist_info: Optional[ArtistInfo] = field(
        default=None, metadata=config(field_name="artistInfo2")
    )

    albums: Optional[AlbumList2] = field(
        default=None, metadata=config(field_name="albumList2")
    )
    album: Optional[Album] = None

    directory: Optional[Directory] = None

    genres: Optional[Genres] = None

    indexes: Optional[Indexes] = None

    playlist: Optional[Playlist] = None
    playlists: Optional[Playlists] = None

    play_queue: Optional[PlayQueue] = field(
        default=None, metadata=config(field_name="playQueue")
    )

    song: Optional[Song] = None

    search_result: Optional[SearchResult3] = field(
        default=None, metadata=config(field_name="searchResult3")
    )



from typing import Optional

from gi.repository import GdkPixbuf, Gtk


class SpinnerImage(Gtk.Overlay):
    def __init__(
        self,
        loading: bool = True,
        image_name: str = None,
        spinner_name: str = None,
        image_size: int = None,
        **kwargs,
    ):
        """An image with a loading overlay."""
        Gtk.Overlay.__init__(self)
        self.image_size = image_size
        self.filename: Optional[str] = None

        self.image = Gtk.Image(name=image_name, **kwargs)
        self.add(self.image)

        self.spinner = Gtk.Spinner(
            name=spinner_name,
            active=loading,
            halign=Gtk.Align.CENTER,
            valign=Gtk.Align.CENTER,
        )
        self.add_overlay(self.spinner)

    def set_from_file(self, filename: Optional[str]):
        """Set the image to the given filename."""
        if filename == "":
            filename = None
        self.filename = filename
        if self.image_size is not None and filename:
            pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_scale(
                filename, self.image_size, self.image_size, True
            )
            self.image.set_from_pixbuf(pixbuf)
        else:
            self.image.set_from_file(filename)

    def set_loading(self, loading_status: bool):
        if loading_status:
            self.spinner.start()
            self.spinner.show()
        else:
            self.spinner.stop()
            self.spinner.hide()

    def set_image_size(self, size: int):
        self.image_size = size
        self.set_from_file(self.filename)



import hashlib
import itertools
import logging
import os
import random
import tempfile
import threading
from concurrent.futures import Future, ThreadPoolExecutor
from dataclasses import dataclass
from datetime import timedelta
from enum import Enum
from functools import partial
from pathlib import Path
from time import sleep
from typing import (
    Any,
    Callable,
    cast,
    Dict,
    Generic,
    Iterable,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
    TypeVar,
    Union,
)

import requests

from sublime_music.config import ProviderConfiguration

from .adapter_base import (
    Adapter,
    AlbumSearchQuery,
    CacheMissError,
    CachingAdapter,
    SongCacheStatus,
)
from .api_objects import (
    Album,
    Artist,
    Directory,
    Genre,
    Playlist,
    PlayQueue,
    SearchResult,
    Song,
)
from .filesystem import FilesystemAdapter
from .subsonic import SubsonicAdapter
from ..util import resolve_path

REQUEST_DELAY: Optional[Tuple[float, float]] = None
if delay_str := os.environ.get("REQUEST_DELAY"):
    if "," in delay_str:
        high, low = map(float, delay_str.split(","))
        REQUEST_DELAY = (high, low)
    else:
        REQUEST_DELAY = (float(delay_str), float(delay_str))

NETWORK_ALWAYS_ERROR: bool = False
if os.environ.get("NETWORK_ALWAYS_ERROR"):
    NETWORK_ALWAYS_ERROR = True

DOWNLOAD_BLOCK_DELAY: Optional[float] = None
if delay_str := os.environ.get("DOWNLOAD_BLOCK_DELAY"):
    DOWNLOAD_BLOCK_DELAY = float(delay_str)

T = TypeVar("T")


class Result(Generic[T]):
    """
    A result from a :class:`AdapterManager` function. This is effectively a wrapper
    around a :class:`concurrent.futures.Future`, but it resolves immediately if the data
    already exists.
    """

    _data: Optional[T] = None
    _future: Optional[Future] = None
    _default_value: Optional[T] = None
    _on_cancel: Optional[Callable[[], None]] = None
    _cancelled = False

    def __init__(
        self,
        data_resolver: Union[T, Callable[[], T]],
        *args,
        is_download: bool = False,
        default_value: T = None,
        on_cancel: Callable[[], None] = None,
    ):
        """
        Creates a :class:`Result` object.

        :param data_resolver: the actual data, or a function that will return the actual
            data. If the latter, the function will be executed by the thread pool.
        :param is_download: whether or not this result requires a file download. If it
            does, then it uses a separate executor.
        """
        if callable(data_resolver):
            if is_download:
                self._future = AdapterManager.download_executor.submit(
                    data_resolver, *args
                )
            else:
                self._future = AdapterManager.executor.submit(data_resolver, *args)
            self._future.add_done_callback(self._on_future_complete)
        else:
            self._data = data_resolver

        self._default_value = default_value
        self._on_cancel = on_cancel

    def _on_future_complete(self, future: Future):
        try:
            self._data = future.result()
        except Exception as e:
            if self._default_value:
                self._data = self._default_value
            else:
                raise e

    def result(self) -> T:
        """
        Retrieve the actual data. If the data exists already, then return it, otherwise,
        blocking-wait on the future's result.
        """
        try:
            if self._data is not None:
                return self._data
            if self._future is not None:
                return self._future.result()

            assert 0, "AdapterManager.Result had neither _data nor _future member!"
        except Exception as e:
            if self._default_value:
                return self._default_value
            else:
                raise e

    def add_done_callback(self, fn: Callable, *args):
        """Attaches the callable ``fn`` to the future."""
        if self._future is not None:
            self._future.add_done_callback(fn, *args)
        else:
            # Run the function immediately if it's not a future.
            fn(self, *args)

    def cancel(self) -> bool:
        """Cancel the future, or do nothing if the data already exists."""
        if self._on_cancel:
            self._on_cancel()
        if self._future is not None:
            return self._future.cancel()
        self._cancelled = True
        return True

    def cancelled(self) -> bool:
        return self._cancelled

    @property
    def data_is_available(self) -> bool:
        """
        Whether or not the data is available at the current moment. This can be used to
        determine whether or not the UI needs to put the callback into a
        :class:`GLib.idle_add` call.
        """
        return self._data is not None


@dataclass
class DownloadProgress:
    class Type(Enum):
        QUEUED = 0
        PROGRESS = 1
        DONE = 2
        CANCELLED = 3
        ERROR = 4

    type: Type
    total_bytes: Optional[int] = None
    current_bytes: Optional[int] = None
    exception: Optional[Exception] = None

    @property
    def progress_fraction(self) -> Optional[float]:
        if self.current_bytes is None or self.total_bytes is None:
            return None
        return self.current_bytes / self.total_bytes


class AdapterManager:
    available_adapters: Set[Any] = {FilesystemAdapter, SubsonicAdapter}
    current_download_ids: Set[str] = set()
    download_set_lock = threading.Lock()
    executor: ThreadPoolExecutor = ThreadPoolExecutor()
    download_executor: ThreadPoolExecutor = ThreadPoolExecutor()
    is_shutting_down: bool = False
    _offline_mode: bool = False

    _song_download_jobs: Dict[str, Result[str]] = {}
    _cancelled_song_ids: Set[str] = set()

    @dataclass
    class _AdapterManagerInternal:
        ground_truth_adapter: Adapter
        on_song_download_progress: Callable[[Any, str, DownloadProgress], None]
        caching_adapter: Optional[CachingAdapter] = None
        concurrent_download_limit: int = 5

        def __post_init__(self):
            self._download_dir = tempfile.TemporaryDirectory()
            self.download_path = Path(self._download_dir.name)
            self.download_limiter_semaphore = threading.Semaphore(
                self.concurrent_download_limit
            )

        def song_download_progress(self, file_id: str, progress: DownloadProgress):
            self.on_song_download_progress(file_id, progress)

        def shutdown(self):
            self.ground_truth_adapter.shutdown()
            if self.caching_adapter:
                self.caching_adapter.shutdown()
            self._download_dir.cleanup()

    _instance: Optional[_AdapterManagerInternal] = None

    def __init__(self):
        """
        This should not ever be called. You should only ever use the static methods on
        this class.
        """
        raise Exception(
            "Do not instantiate the AdapterManager. Only use the static methods on the class."  # noqa: 512
        )

    @staticmethod
    def initial_sync() -> Result[None]:
        assert AdapterManager._instance
        return Result(AdapterManager._instance.ground_truth_adapter.initial_sync)

    @staticmethod
    def ground_truth_adapter_is_networked() -> bool:
        assert AdapterManager._instance
        return AdapterManager._instance.ground_truth_adapter.is_networked

    @staticmethod
    def get_ping_status() -> bool:
        assert AdapterManager._instance
        return AdapterManager._instance.ground_truth_adapter.ping_status

    @staticmethod
    def shutdown():
        logging.info("AdapterManager shutdown start")
        AdapterManager.is_shutting_down = True
        for _, job in AdapterManager._song_download_jobs.items():
            job.cancel()

        AdapterManager.executor.shutdown()
        AdapterManager.download_executor.shutdown()
        if AdapterManager._instance:
            AdapterManager._instance.shutdown()

        logging.info("AdapterManager shutdown complete")

    @staticmethod
    def reset(
        config: Any,
        on_song_download_progress: Callable[[Any, str, DownloadProgress], None],
    ):
        from sublime_music.config import AppConfiguration

        assert isinstance(config, AppConfiguration)

        # First, shutdown the current one...
        if AdapterManager._instance:
            AdapterManager._instance.shutdown()

        AdapterManager._offline_mode = config.offline_mode

        assert config.provider is not None
        assert isinstance(config.provider, ProviderConfiguration)
        assert config.cache_location

        source_data_dir = config.cache_location.joinpath(config.provider.id)
        source_data_dir.joinpath("g").mkdir(parents=True, exist_ok=True)
        source_data_dir.joinpath("c").mkdir(parents=True, exist_ok=True)

        ground_truth_adapter = config.provider.ground_truth_adapter_type(
            config.provider.ground_truth_adapter_config, source_data_dir.joinpath("g")
        )

        caching_adapter = None
        if (
            caching_adapter_type := config.provider.caching_adapter_type
        ) and config.provider.ground_truth_adapter_type.can_be_cached:
            caching_adapter = caching_adapter_type(
                config.provider.caching_adapter_config,
                source_data_dir.joinpath("c"),
                is_cache=True,
            )

        AdapterManager._instance = AdapterManager._AdapterManagerInternal(
            ground_truth_adapter,
            on_song_download_progress,
            caching_adapter=caching_adapter,
            concurrent_download_limit=config.concurrent_download_limit,
        )

    @staticmethod
    def on_offline_mode_change(offline_mode: bool):
        AdapterManager._offline_mode = offline_mode
        if (instance := AdapterManager._instance) and (
            (ground_truth_adapter := instance.ground_truth_adapter).is_networked
        ):
            ground_truth_adapter.on_offline_mode_change(offline_mode)

    # Data Helper Methods
    # ==================================================================================
    TAdapter = TypeVar("TAdapter", bound=Adapter)

    @staticmethod
    def _adapter_can_do(adapter: TAdapter, action_name: str) -> bool:
        return adapter is not None and getattr(adapter, f"can_{action_name}", False)

    @staticmethod
    def _ground_truth_can_do(action_name: str) -> bool:
        if not AdapterManager._instance:
            return False
        return AdapterManager._adapter_can_do(
            AdapterManager._instance.ground_truth_adapter, action_name
        )

    @staticmethod
    def _can_use_cache(force: bool, action_name: str) -> bool:
        if force:
            return False
        return (
            AdapterManager._instance is not None
            and AdapterManager._instance.caching_adapter is not None
            and AdapterManager._adapter_can_do(
                AdapterManager._instance.caching_adapter, action_name
            )
        )

    @staticmethod
    def _create_ground_truth_result(
        function_name: str,
        *params: Any,
        before_download: Callable[[], None] = None,
        partial_data: Any = None,
        **kwargs,
    ) -> Result:
        """
        Creates a Result using the given ``function_name`` on the ground truth adapter.
        """

        def future_fn() -> Any:
            assert AdapterManager._instance
            if (
                AdapterManager._offline_mode
                and AdapterManager._instance.ground_truth_adapter.is_networked
            ):
                raise CacheMissError(partial_data=partial_data)

            if before_download:
                before_download()
            fn = getattr(AdapterManager._instance.ground_truth_adapter, function_name)
            try:
                return fn(*params, **kwargs)
            except Exception as e:
                raise CacheMissError(partial_data=partial_data) from e

        return Result(future_fn)

    @staticmethod
    def _create_download_result(
        uri: str,
        id: str,
        before_download: Callable[[], None] = None,
        expected_size: int = None,
        **result_args,
    ) -> Result[str]:
        """
        Create a function to download the given URI to a temporary file, and return the
        filename. The returned function will spin-loop if the resource is already being
        downloaded to prevent multiple requests for the same download.
        """
        download_cancelled = False

        def download_fn() -> str:
            assert AdapterManager._instance
            download_tmp_filename = AdapterManager._instance.download_path.joinpath(
                hashlib.sha1(bytes(uri, "utf8")).hexdigest()
            )

            resource_downloading = False
            with AdapterManager.download_set_lock:
                if id in AdapterManager.current_download_ids:
                    resource_downloading = True
                AdapterManager.current_download_ids.add(id)

            if before_download:
                before_download()

            expected_size_exists = expected_size is not None
            if expected_size_exists:
                AdapterManager._instance.song_download_progress(
                    id,
                    DownloadProgress(
                        DownloadProgress.Type.PROGRESS,
                        total_bytes=expected_size,
                        current_bytes=0,
                    ),
                )

            # TODO (#122): figure out how to retry if the other request failed.
            if resource_downloading:
                logging.info(f"{uri} already being downloaded.")

                # The resource is already being downloaded. Busy loop until
                # it has completed. Then, just return the path to the
                # resource.
                t = 0.0
                while id in AdapterManager.current_download_ids and t < 20:
                    sleep(0.2)
                    t += 0.2
                    # TODO (#122): handle the timeout
            else:
                logging.info(f"{uri} not found. Downloading...")
                try:
                    if REQUEST_DELAY is not None:
                        delay = random.uniform(*REQUEST_DELAY)
                        logging.info(
                            f"REQUEST_DELAY enabled. Pausing for {delay} seconds"
                        )
                        sleep(delay)

                    if NETWORK_ALWAYS_ERROR:
                        raise Exception("NETWORK_ALWAYS_ERROR enabled")

                    # Wait 10 seconds to connect to the server and start downloading.
                    # Then, for each of the blocks, give 5 seconds to download (which
                    # should be more than enough for 1 KiB).
                    request = requests.get(uri, stream=True, timeout=(10, 5))
                    if "json" in request.headers.get("Content-Type", ""):
                        raise Exception("Didn't expect JSON!")

                    total_size = int(request.headers.get("Content-Length", 0))
                    if expected_size_exists:
                        if total_size != expected_size:
                            raise Exception(
                                f"Download content size ({total_size})is not the "
                                f"expected size ({expected_size})."
                            )

                    block_size = 1024  # 1 KiB
                    total_consumed = 0

                    with open(download_tmp_filename, "wb+") as f:
                        for i, data in enumerate(request.iter_content(block_size)):
                            total_consumed += len(data)
                            f.write(data)

                            if download_cancelled:
                                AdapterManager._instance.song_download_progress(
                                    id,
                                    DownloadProgress(DownloadProgress.Type.CANCELLED),
                                )
                                raise Exception("Download Cancelled")

                            if i % 100 == 0:
                                # Only delay (if configured) and update the progress UI
                                # every 100 KiB.
                                if DOWNLOAD_BLOCK_DELAY is not None:
                                    sleep(DOWNLOAD_BLOCK_DELAY)

                                if expected_size_exists:
                                    AdapterManager._instance.song_download_progress(
                                        id,
                                        DownloadProgress(
                                            DownloadProgress.Type.PROGRESS,
                                            total_bytes=total_size,
                                            current_bytes=total_consumed,
                                        ),
                                    )

                    # Everything succeeded.
                    if expected_size_exists:
                        AdapterManager._instance.song_download_progress(
                            id,
                            DownloadProgress(DownloadProgress.Type.DONE),
                        )
                except Exception as e:
                    if expected_size_exists and not download_cancelled:
                        # Something failed. Post an error.
                        AdapterManager._instance.song_download_progress(
                            id,
                            DownloadProgress(DownloadProgress.Type.ERROR, exception=e),
                        )
                    # Re-raise the exception so that we can actually handle it.
                    raise
                finally:
                    # Always release the download set lock, even if there's an error.
                    with AdapterManager.download_set_lock:
                        AdapterManager.current_download_ids.discard(id)

            logging.info(f"{uri} downloaded. Returning.")
            return str(download_tmp_filename)

        def on_download_cancel():
            nonlocal download_cancelled
            download_cancelled = True

        return Result(
            download_fn, is_download=True, on_cancel=on_download_cancel, **result_args
        )

    @staticmethod
    def _create_caching_done_callback(
        cache_key: CachingAdapter.CachedDataKey, param: Optional[str]
    ) -> Callable[[Result], None]:
        """
        Create a function to let the caching_adapter ingest new data.

        :param cache_key: the cache key to ingest.
        :param params: the parameters to uniquely identify the cached item.
        """

        def future_finished(f: Result):
            assert AdapterManager._instance
            assert AdapterManager._instance.caching_adapter
            AdapterManager._instance.caching_adapter.ingest_new_data(
                cache_key, param, f.result()
            )

        return future_finished

    @staticmethod
    def get_supported_artist_query_types() -> Set[AlbumSearchQuery.Type]:
        assert AdapterManager._instance
        return (
            AdapterManager._instance.ground_truth_adapter.supported_artist_query_types
        )

    R = TypeVar("R")

    @staticmethod
    def _get_from_cache_or_ground_truth(
        function_name: str,
        param: Optional[Union[str, AlbumSearchQuery]],
        cache_key: CachingAdapter.CachedDataKey = None,
        before_download: Callable[[], None] = None,
        use_ground_truth_adapter: bool = False,
        allow_download: bool = True,
        on_result_finished: Callable[[Result], None] = None,
        **kwargs: Any,
    ) -> Result[R]:
        """
        Get data from one of the adapters.

        :param function_name: The function to call on the adapter.
        :param param: The parameter to pass to the adapter function (also used for the
            cache parameter to uniquely identify the request).
        :param cache_key: The cache key to use to invalidate caches and ingest caches.
        :param before_download: Function to call before doing a network request.
        :param allow_download: Whether or not to allow a network request to retrieve the
            data.
        :param on_result_finished: A function to run after the result received from the
            ground truth adapter. (Has no effect if the result is from the caching
            adapter.)
        :param kwargs: The keyword arguments to pass to the adapter function.
        """
        assert AdapterManager._instance
        logging.info(f"START: {function_name}")
        partial_data = None
        if AdapterManager._can_use_cache(use_ground_truth_adapter, function_name):
            assert (caching_adapter := AdapterManager._instance.caching_adapter)
            try:
                logging.info(f"END: {function_name}: serving from cache")
                if param is None:
                    return Result(getattr(caching_adapter, function_name)(**kwargs))
                return Result(getattr(caching_adapter, function_name)(param, **kwargs))
            except CacheMissError as e:
                partial_data = e.partial_data
                logging.info(f"Cache Miss on {function_name}.")
            except Exception:
                logging.exception(f"Error on {function_name} retrieving from cache.")

        param_str = param.strhash() if isinstance(param, AlbumSearchQuery) else param
        if (
            cache_key
            and AdapterManager._instance.caching_adapter
            and use_ground_truth_adapter
        ):
            AdapterManager._instance.caching_adapter.invalidate_data(
                cache_key, param_str
            )

        if (
            not allow_download
            and AdapterManager._instance.ground_truth_adapter.is_networked
        ) or not AdapterManager._ground_truth_can_do(function_name):
            logging.info(f"END: NO DOWNLOAD: {function_name}")

            def cache_miss_result():
                raise CacheMissError(partial_data=partial_data)

            return Result(cache_miss_result)

        result: Result[AdapterManager.R] = AdapterManager._create_ground_truth_result(
            function_name,
            *((param,) if param is not None else ()),
            before_download=before_download,
            partial_data=partial_data,
            **kwargs,
        )

        if AdapterManager._instance.caching_adapter:
            if cache_key:
                result.add_done_callback(
                    AdapterManager._create_caching_done_callback(cache_key, param_str)
                )

            if on_result_finished:
                result.add_done_callback(on_result_finished)

        logging.info(f"END: {function_name}")
        logging.debug(result)
        return result

    # Usage and Availability Properties
    # ==================================================================================
    @staticmethod
    def can_get_playlists() -> bool:
        return AdapterManager._ground_truth_can_do("get_playlists")

    @staticmethod
    def can_get_playlist_details() -> bool:
        return AdapterManager._ground_truth_can_do("get_playlist_details")

    @staticmethod
    def can_create_playlist() -> bool:
        return AdapterManager._ground_truth_can_do("create_playlist")

    @staticmethod
    def can_update_playlist() -> bool:
        return AdapterManager._ground_truth_can_do("update_playlist")

    @staticmethod
    def can_delete_playlist() -> bool:
        return AdapterManager._ground_truth_can_do("delete_playlist")

    @staticmethod
    def can_get_song_file_uri() -> bool:
        return AdapterManager._ground_truth_can_do("get_song_file_uri")

    @staticmethod
    def can_get_song_stream_uri() -> bool:
        return AdapterManager._ground_truth_can_do("get_song_stream_uri")

    @staticmethod
    def can_batch_download_songs() -> bool:
        # We can only download from the ground truth adapter.
        return AdapterManager._ground_truth_can_do("get_song_file_uri")

    @staticmethod
    def can_get_genres() -> bool:
        return AdapterManager._ground_truth_can_do("get_genres")

    @staticmethod
    def can_scrobble_song() -> bool:
        return AdapterManager._ground_truth_can_do("scrobble_song")

    @staticmethod
    def can_get_artists() -> bool:
        return AdapterManager._ground_truth_can_do("get_artists")

    @staticmethod
    def can_get_artist() -> bool:
        return AdapterManager._ground_truth_can_do("get_artist")

    @staticmethod
    def can_get_directory() -> bool:
        return AdapterManager._ground_truth_can_do("get_directory")

    @staticmethod
    def can_get_play_queue() -> bool:
        return AdapterManager._ground_truth_can_do("get_play_queue")

    @staticmethod
    def can_save_play_queue() -> bool:
        return AdapterManager._ground_truth_can_do("save_play_queue")

    @staticmethod
    def can_search() -> bool:
        return AdapterManager._ground_truth_can_do("search")

    # Data Retrieval Methods
    # ==================================================================================
    @staticmethod
    def get_playlists(
        before_download: Callable[[], None] = lambda: None,
        force: bool = False,  # TODO (#202): rename to use_ground_truth_adapter?
        allow_download: bool = True,
    ) -> Result[Sequence[Playlist]]:
        return AdapterManager._get_from_cache_or_ground_truth(
            "get_playlists",
            None,
            cache_key=CachingAdapter.CachedDataKey.PLAYLISTS,
            before_download=before_download,
            use_ground_truth_adapter=force,
            allow_download=allow_download,
        )

    @staticmethod
    def get_playlist_details(
        playlist_id: str,
        before_download: Callable[[], None] = lambda: None,
        force: bool = False,  # TODO (#202): rename to use_ground_truth_adapter?
        allow_download: bool = True,
    ) -> Result[Playlist]:
        return AdapterManager._get_from_cache_or_ground_truth(
            "get_playlist_details",
            playlist_id,
            cache_key=CachingAdapter.CachedDataKey.PLAYLIST_DETAILS,
            before_download=before_download,
            use_ground_truth_adapter=force,
            allow_download=allow_download,
        )

    @staticmethod
    def create_playlist(
        name: str, songs: Sequence[Song] = None
    ) -> Result[Optional[Playlist]]:
        def on_result_finished(f: Result[Optional[Playlist]]):
            assert AdapterManager._instance
            assert AdapterManager._instance.caching_adapter
            if playlist := f.result():
                AdapterManager._instance.caching_adapter.ingest_new_data(
                    CachingAdapter.CachedDataKey.PLAYLIST_DETAILS,
                    playlist.id,
                    playlist,
                )
            else:
                AdapterManager._instance.caching_adapter.invalidate_data(
                    CachingAdapter.CachedDataKey.PLAYLISTS, None
                )

        return AdapterManager._get_from_cache_or_ground_truth(
            "create_playlist",
            name,
            songs=songs,
            on_result_finished=on_result_finished,
            use_ground_truth_adapter=True,
        )

    @staticmethod
    def update_playlist(
        playlist_id: str,
        name: str = None,
        comment: str = None,
        public: bool = False,
        song_ids: Sequence[str] = None,
        append_song_ids: Sequence[str] = None,
        before_download: Callable[[], None] = lambda: None,
    ) -> Result[Playlist]:
        return AdapterManager._get_from_cache_or_ground_truth(
            "update_playlist",
            playlist_id,
            name=name,
            comment=comment,
            public=public,
            song_ids=song_ids,
            append_song_ids=append_song_ids,
            before_download=before_download,
            use_ground_truth_adapter=True,
            cache_key=CachingAdapter.CachedDataKey.PLAYLIST_DETAILS,
        )

    @staticmethod
    def delete_playlist(playlist_id: str):
        assert AdapterManager._instance
        ground_truth_adapter = AdapterManager._instance.ground_truth_adapter
        if AdapterManager._offline_mode and ground_truth_adapter.is_networked:
            raise AssertionError(
                "You should never call delete_playlist in offline mode"
            )

        # TODO (#190): make non-blocking?
        ground_truth_adapter.delete_playlist(playlist_id)

        if AdapterManager._instance.caching_adapter:
            AdapterManager._instance.caching_adapter.delete_data(
                CachingAdapter.CachedDataKey.PLAYLIST_DETAILS, playlist_id
            )

    @staticmethod
    def _get_networked_scheme() -> str:
        assert AdapterManager._instance
        networked_scheme_priority = ("https", "http")
        return sorted(
            AdapterManager._instance.ground_truth_adapter.supported_schemes,
            key=lambda s: networked_scheme_priority.index(s),
        )[0]

    @staticmethod
    def get_cover_art_uri(
        cover_art_id: Optional[str],
        scheme: str,
        size: int = 300,
        before_download: Callable[[], None] = None,
        force: bool = False,
        allow_download: bool = True,
    ) -> Result[str]:
        existing_filename = str(resolve_path("adapters/images/default-album-art.png"))
        if (
            not AdapterManager._ground_truth_can_do("get_cover_art_uri")
            or not cover_art_id
        ):
            return Result(existing_filename if scheme == "file" else "")

        assert AdapterManager._instance
        supported_schemes = (
            AdapterManager._instance.ground_truth_adapter.supported_schemes
        )

        # If the scheme is supported natively, then return it.
        if scheme in supported_schemes:
            uri = AdapterManager._instance.ground_truth_adapter.get_cover_art_uri(
                cover_art_id, scheme, size=size
            )
            return Result(uri)

        # If the scheme is "file", then we may need to try to download.
        if scheme == "file" and (
            "http" in supported_schemes or "https" in supported_schemes
        ):
            if AdapterManager._can_use_cache(force, "get_cover_art_uri"):
                assert AdapterManager._instance.caching_adapter
                try:
                    return Result(
                        AdapterManager._instance.caching_adapter.get_cover_art_uri(
                            cover_art_id, "file", size=size
                        )
                    )
                except CacheMissError as e:
                    if e.partial_data is not None:
                        existing_filename = cast(str, e.partial_data)
                    logging.info("Cache Miss on get_cover_art_uri.")
                except Exception:
                    logging.exception(
                        "Error on get_cover_art_uri retrieving from cache."
                    )

            # If we are forcing, invalidate the existing cached data.
            if AdapterManager._instance.caching_adapter and force:
                AdapterManager._instance.caching_adapter.invalidate_data(
                    CachingAdapter.CachedDataKey.COVER_ART_FILE, cover_art_id
                )

            if not allow_download or (
                AdapterManager._offline_mode
                and AdapterManager._instance.ground_truth_adapter.is_networked
            ):
                return Result(existing_filename)

            # Create a download result.
            future = AdapterManager._create_download_result(
                AdapterManager._instance.ground_truth_adapter.get_cover_art_uri(
                    cover_art_id,
                    AdapterManager._get_networked_scheme(),
                    size=size,
                ),
                cover_art_id,
                before_download,
                default_value=existing_filename,
            )

            if AdapterManager._instance.caching_adapter:
                future.add_done_callback(
                    AdapterManager._create_caching_done_callback(
                        CachingAdapter.CachedDataKey.COVER_ART_FILE, cover_art_id
                    )
                )

            return future

        return Result("")

    @staticmethod
    def get_song_file_uri(song: Song) -> str:
        assert AdapterManager._instance
        cached_song_filename = None
        if AdapterManager._can_use_cache(False, "get_song_file_uri"):
            assert (caching_adapter := AdapterManager._instance.caching_adapter)
            try:
                if "file" not in caching_adapter.supported_schemes:
                    raise Exception("file not a supported scheme")

                return caching_adapter.get_song_file_uri(song.id, "file")
            except CacheMissError as e:
                if e.partial_data is not None:
                    cached_song_filename = cast(str, e.partial_data)
                logging.info("Cache Miss on get_song_file_uri.")
            except Exception:
                logging.exception("Error on get_song_file_uri retrieving from cache.")

        ground_truth_adapter = AdapterManager._instance.ground_truth_adapter
        if (
            not AdapterManager._ground_truth_can_do("get_song_file_uri")
            or (ground_truth_adapter.is_networked and AdapterManager._offline_mode)
            or ("file" not in ground_truth_adapter.supported_schemes)
        ):
            raise CacheMissError(partial_data=cached_song_filename)

        return ground_truth_adapter.get_song_file_uri(song.id, "file")

    @staticmethod
    def get_song_stream_uri(song: Song) -> str:
        assert AdapterManager._instance
        if not AdapterManager._ground_truth_can_do("get_song_stream_uri"):
            raise Exception(f"Can't stream song '{song.title}'.")
        return AdapterManager._instance.ground_truth_adapter.get_song_stream_uri(
            song.id
        )

    @staticmethod
    def batch_download_songs(
        song_ids: Sequence[str],
        before_download: Callable[[str], None],
        on_song_download_complete: Callable[[str], None],
        one_at_a_time: bool = False,
        delay: float = 0.0,
    ) -> Result[None]:
        assert AdapterManager._instance
        if (
            AdapterManager._offline_mode
            and AdapterManager._instance.ground_truth_adapter.is_networked
        ):
            raise AssertionError(
                "You should never call batch_download_songs in offline mode"
            )

        # This only really makes sense if we have a caching_adapter.
        if not AdapterManager._instance.caching_adapter:
            return Result(None)

        cancelled = False
        AdapterManager._cancelled_song_ids -= set(song_ids)

        def do_download_song(song_id: str) -> Result:
            assert AdapterManager._instance
            assert AdapterManager._instance.caching_adapter

            if (
                AdapterManager.is_shutting_down
                or AdapterManager._offline_mode
                or cancelled
                or song_id in AdapterManager._cancelled_song_ids
            ):
                AdapterManager._instance.download_limiter_semaphore.release()
                AdapterManager._instance.song_download_progress(
                    song_id,
                    DownloadProgress(DownloadProgress.Type.CANCELLED),
                )
                return Result("", is_download=True)

            logging.info(f"Downloading {song_id}")

            # Download the actual song file.
            try:
                # If the song file is already cached, just indicate done immediately.
                AdapterManager._instance.caching_adapter.get_song_file_uri(
                    song_id, "file"
                )
                AdapterManager._instance.download_limiter_semaphore.release()
                AdapterManager._instance.song_download_progress(
                    song_id,
                    DownloadProgress(DownloadProgress.Type.DONE),
                )
                return Result("", is_download=True)
            except CacheMissError:
                # The song is not already cached.
                if before_download:
                    before_download(song_id)

                song = AdapterManager.get_song_details(song_id).result()

                # Download the song.
                song_tmp_filename_result: Result[
                    str
                ] = AdapterManager._create_download_result(
                    AdapterManager._instance.ground_truth_adapter.get_song_file_uri(
                        song_id, AdapterManager._get_networked_scheme()
                    ),
                    song_id,
                    lambda: before_download(song_id),
                    expected_size=song.size,
                )

                def on_download_done(f: Result):
                    assert AdapterManager._instance
                    assert AdapterManager._instance.caching_adapter
                    AdapterManager._instance.download_limiter_semaphore.release()

                    try:
                        AdapterManager._instance.caching_adapter.ingest_new_data(
                            CachingAdapter.CachedDataKey.SONG_FILE,
                            song_id,
                            (None, f.result(), None),
                        )
                    finally:
                        if AdapterManager._song_download_jobs.get(song_id):
                            del AdapterManager._song_download_jobs[song_id]

                        on_song_download_complete(song_id)

                song_tmp_filename_result.add_done_callback(on_download_done)
                AdapterManager._song_download_jobs[song_id] = song_tmp_filename_result
                return song_tmp_filename_result

        def do_batch_download_songs():
            sleep(delay)
            if (
                AdapterManager.is_shutting_down
                or AdapterManager._offline_mode
                or cancelled
            ):
                return

            # Alert the UI that the downloads are queued.
            for song_id in song_ids:
                # Everything succeeded.
                AdapterManager._instance.song_download_progress(
                    song_id,
                    DownloadProgress(DownloadProgress.Type.QUEUED),
                )

            for song_id in song_ids:
                # Only allow a certain number of songs to be downloaded
                # simultaneously.
                AdapterManager._instance.download_limiter_semaphore.acquire()

                result = do_download_song(song_id)

                if one_at_a_time:
                    # Wait the file to download.
                    result.result()

        def on_cancel():
            nonlocal cancelled
            cancelled = True

            # Cancel the individual song downloads
            AdapterManager.cancel_download_songs(song_ids)

            # Alert the UI that the downloads are cancelled.
            for song_id in song_ids:
                AdapterManager._instance.song_download_progress(
                    song_id,
                    DownloadProgress(DownloadProgress.Type.CANCELLED),
                )

        return Result(do_batch_download_songs, is_download=True, on_cancel=on_cancel)

    @staticmethod
    def cancel_download_songs(song_ids: Sequence[str]):
        assert AdapterManager._instance
        AdapterManager._cancelled_song_ids = AdapterManager._cancelled_song_ids.union(
            set(song_ids)
        )
        for song_id in song_ids:
            AdapterManager._instance.song_download_progress(
                song_id,
                DownloadProgress(DownloadProgress.Type.CANCELLED),
            )
            if AdapterManager._song_download_jobs.get(song_id):
                AdapterManager._song_download_jobs[song_id].cancel()
                del AdapterManager._song_download_jobs[song_id]

    @staticmethod
    def batch_permanently_cache_songs(
        song_ids: Sequence[str],
        before_download: Callable[[str], None],
        on_song_download_complete: Callable[[str], None],
    ) -> Result[None]:
        assert AdapterManager._instance
        # This only really makes sense if we have a caching_adapter.
        if not AdapterManager._instance.caching_adapter:
            return Result(None)
        # TODO (#74): actually implement this
        raise NotImplementedError()

    @staticmethod
    def batch_delete_cached_songs(
        song_ids: Sequence[str], on_song_delete: Callable[[str], None]
    ):
        assert AdapterManager._instance

        # This only really makes sense if we have a caching_adapter.
        if not AdapterManager._instance.caching_adapter:
            return

        for song_id in song_ids:
            song = AdapterManager.get_song_details(song_id).result()
            AdapterManager._instance.caching_adapter.delete_data(
                CachingAdapter.CachedDataKey.SONG_FILE, song.id
            )
            on_song_delete(song_id)

    @staticmethod
    def get_song_details(
        song_id: str,
        allow_download: bool = True,
        before_download: Callable[[], None] = lambda: None,
        force: bool = False,
    ) -> Result[Song]:
        return AdapterManager._get_from_cache_or_ground_truth(
            "get_song_details",
            song_id,
            allow_download=allow_download,
            before_download=before_download,
            use_ground_truth_adapter=force,
            cache_key=CachingAdapter.CachedDataKey.SONG,
        )

    @staticmethod
    def get_genres(force: bool = False) -> Result[Sequence[Genre]]:
        return AdapterManager._get_from_cache_or_ground_truth(
            "get_genres",
            None,
            use_ground_truth_adapter=force,
            cache_key=CachingAdapter.CachedDataKey.GENRES,
        )

    @staticmethod
    def scrobble_song(song: Song):
        assert AdapterManager._instance
        AdapterManager._create_ground_truth_result("scrobble_song", song)

    @staticmethod
    def get_artists(
        force: bool = False, before_download: Callable[[], None] = lambda: None
    ) -> Result[Sequence[Artist]]:
        def do_get_artists() -> Sequence[Artist]:
            return AdapterManager.sort_by_ignored_articles(
                AdapterManager._get_from_cache_or_ground_truth(
                    "get_artists",
                    None,
                    use_ground_truth_adapter=force,
                    before_download=before_download,
                    cache_key=CachingAdapter.CachedDataKey.ARTISTS,
                ).result(),
                key=lambda a: a.name,
                use_ground_truth_adapter=force,
            )

        return Result(do_get_artists)

    @staticmethod
    def _get_ignored_articles(use_ground_truth_adapter: bool) -> Set[str]:
        # TODO (#21) get this at first startup.
        if not AdapterManager._ground_truth_can_do("get_ignored_articles"):
            return set()
        try:
            ignored_articles: Set[str] = AdapterManager._get_from_cache_or_ground_truth(
                "get_ignored_articles",
                None,
                use_ground_truth_adapter=use_ground_truth_adapter,
                cache_key=CachingAdapter.CachedDataKey.IGNORED_ARTICLES,
            ).result()
            return set(map(str.lower, ignored_articles))
        except Exception:
            logging.exception("Failed to retrieve ignored_articles")
            return set()

    @staticmethod
    def _strip_ignored_articles(
        use_ground_truth_adapter: bool, ignored_articles: Set[str], string: str
    ) -> str:
        first_word, *rest = string.split(maxsplit=1)
        if first_word in ignored_articles:
            return rest[0]
        return string

    _S = TypeVar("_S")

    @staticmethod
    def sort_by_ignored_articles(
        it: Iterable[_S],
        key: Callable[[_S], str],
        use_ground_truth_adapter: bool = False,
    ) -> List[_S]:
        ignored_articles = AdapterManager._get_ignored_articles(
            use_ground_truth_adapter
        )
        strip_fn = partial(
            AdapterManager._strip_ignored_articles,
            use_ground_truth_adapter,
            ignored_articles,
        )
        return sorted(it, key=lambda x: strip_fn(key(x).lower()))

    @staticmethod
    def get_artist(
        artist_id: str,
        before_download: Callable[[], None] = lambda: None,
        force: bool = False,
    ) -> Result[Artist]:
        def on_result_finished(f: Result[Artist]):
            if not force:
                return

            assert AdapterManager._instance
            assert AdapterManager._instance.caching_adapter
            if artist := f.result():
                for album in artist.albums or []:
                    AdapterManager._instance.caching_adapter.invalidate_data(
                        CachingAdapter.CachedDataKey.ALBUM, album.id
                    )

        return AdapterManager._get_from_cache_or_ground_truth(
            "get_artist",
            artist_id,
            before_download=before_download,
            use_ground_truth_adapter=force,
            cache_key=CachingAdapter.CachedDataKey.ARTIST,
            on_result_finished=on_result_finished,
        )

    # Albums
    @staticmethod
    def get_albums(
        query: AlbumSearchQuery,
        sort_direction: str = "ascending",
        before_download: Callable[[], None] = lambda: None,
        use_ground_truth_adapter: bool = False,
    ) -> Result[Sequence[Album]]:
        return AdapterManager._get_from_cache_or_ground_truth(
            "get_albums",
            query,
            sort_direction=sort_direction,
            cache_key=CachingAdapter.CachedDataKey.ALBUMS,
            before_download=before_download,
            use_ground_truth_adapter=use_ground_truth_adapter,
        )

    @staticmethod
    def get_album(
        album_id: str,
        before_download: Callable[[], None] = lambda: None,
        force: bool = False,
    ) -> Result[Album]:
        return AdapterManager._get_from_cache_or_ground_truth(
            "get_album",
            album_id,
            before_download=before_download,
            use_ground_truth_adapter=force,
            cache_key=CachingAdapter.CachedDataKey.ALBUM,
        )

    # Browse
    @staticmethod
    def get_directory(
        directory_id: str,
        before_download: Callable[[], None] = lambda: None,
        force: bool = False,
    ) -> Result[Directory]:
        def do_get_directory() -> Directory:
            directory: Directory = AdapterManager._get_from_cache_or_ground_truth(
                "get_directory",
                directory_id,
                before_download=before_download,
                use_ground_truth_adapter=force,
                cache_key=CachingAdapter.CachedDataKey.DIRECTORY,
            ).result()
            directory.children = AdapterManager.sort_by_ignored_articles(
                directory.children,
                key=lambda c: cast(Directory, c).name or ""
                if hasattr(c, "name")
                else cast(Song, c).title,
                use_ground_truth_adapter=force,
            )
            return directory

        return Result(do_get_directory)

    # Play Queue
    @staticmethod
    def get_play_queue() -> Result[Optional[PlayQueue]]:
        assert AdapterManager._instance
        return AdapterManager._create_ground_truth_result("get_play_queue")

    @staticmethod
    def save_play_queue(
        song_ids: Sequence[str],
        current_song_index: int = None,
        position: timedelta = None,
    ):
        assert AdapterManager._instance
        AdapterManager._create_ground_truth_result(
            "save_play_queue",
            song_ids,
            current_song_index=current_song_index,
            position=position,
        )

    @staticmethod
    def search(
        query: str,
        search_callback: Callable[[SearchResult], None],
        before_download: Callable[[], None] = lambda: None,
    ) -> Result[bool]:
        if query == "":
            search_callback(SearchResult(""))
            return Result(True)

        before_download()

        # Keep track of if the result is cancelled and if it is, then don't do anything
        # with any results.
        cancelled = False

        # This function actually does the search and calls the search_callback when each
        # of the futures completes. Returns whether or not it was cancelled.
        def do_search() -> bool:
            # Sleep for a little while before returning the local results. They are less
            # expensive to retrieve (but they still incur some overhead due to the GTK
            # UI main loop queue).
            sleep(0.3)
            if cancelled:
                logging.info(f"Cancelled query {query} before caching adapter")
                return True

            assert AdapterManager._instance

            # Caching Adapter Results
            search_result = SearchResult(query)
            if AdapterManager._can_use_cache(False, "search"):
                assert AdapterManager._instance.caching_adapter
                try:
                    logging.info(
                        f"Returning caching adapter search results for '{query}'."
                    )
                    search_result.update(
                        AdapterManager._instance.caching_adapter.search(query)
                    )
                    search_callback(search_result)
                except Exception:
                    logging.exception("Error on caching adapter search")

            if not AdapterManager._ground_truth_can_do("search"):
                return False

            # Wait longer to see if the user types anything else so we don't peg the
            # server with tons of requests.
            sleep(
                1 if AdapterManager._instance.ground_truth_adapter.is_networked else 0.3
            )
            if cancelled:
                logging.info(f"Cancelled query {query} before server results")
                return True

            try:
                ground_truth_search_results = (
                    AdapterManager._instance.ground_truth_adapter.search(  # noqa: E501
                        query
                    )
                )
                search_result.update(ground_truth_search_results)
                search_callback(search_result)
            except Exception:
                logging.exception(
                    "Failed getting search results from server for query '{query}'"
                )

            if AdapterManager._instance.caching_adapter:
                AdapterManager._instance.caching_adapter.ingest_new_data(
                    CachingAdapter.CachedDataKey.SEARCH_RESULTS,
                    None,
                    ground_truth_search_results,
                )

            return False

        # When the future is cancelled (this will happen if a new search is created),
        # set cancelled to True so that the search function can abort.
        def on_cancel():
            nonlocal cancelled
            cancelled = True

        return Result(do_search, on_cancel=on_cancel)

    # Cache Status Methods
    # ==================================================================================
    @staticmethod
    def get_cached_statuses(song_ids: Sequence[str]) -> Sequence[SongCacheStatus]:
        assert AdapterManager._instance
        if not AdapterManager._instance.caching_adapter:
            return list(itertools.repeat(SongCacheStatus.NOT_CACHED, len(song_ids)))

        cached_statuses = AdapterManager._instance.caching_adapter.get_cached_statuses(
            song_ids
        )
        return [
            SongCacheStatus.DOWNLOADING
            if (
                song_id in AdapterManager.current_download_ids
                and song_id not in AdapterManager._cancelled_song_ids
            )
            else cached_statuses[song_id]
            for song_id in song_ids
        ]

    @staticmethod
    def clear_song_cache():
        assert AdapterManager._instance
        if not AdapterManager._instance.caching_adapter:
            return
        AdapterManager._instance.caching_adapter.delete_data(
            CachingAdapter.CachedDataKey.ALL_SONGS, None
        )

    @staticmethod
    def clear_entire_cache():
        assert AdapterManager._instance
        if not AdapterManager._instance.caching_adapter:
            return
        AdapterManager._instance.caching_adapter.delete_data(
            CachingAdapter.CachedDataKey.EVERYTHING, None
        )



import abc
import copy
import hashlib
import uuid
from dataclasses import dataclass
from datetime import timedelta
from enum import Enum
from pathlib import Path
from typing import (
    Any,
    cast,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Set,
    Tuple,
)

from gi.repository import Gtk

try:
    import keyring

    keyring_imported = True
except Exception:
    keyring_imported = False

from .api_objects import (
    Album,
    Artist,
    Directory,
    Genre,
    Playlist,
    PlayQueue,
    SearchResult,
    Song,
)


class SongCacheStatus(Enum):
    """
    Represents the cache state of a given song.

    * :class:`SongCacheStatus.NOT_CACHED` -- indicates that the song is not cached on
      disk.
    * :class:`SongCacheStatus.CACHED` -- indicates that the song is cached on disk.
    * :class:`SongCacheStatus.PERMANENTLY_CACHED` -- indicates that the song is cached
      on disk and will not be deleted when the cache gets too big.
    * :class:`SongCacheStatus.DOWNLOADING` -- indicates that the song is being
      downloaded.
    * :class:`SongCacheStatus.CACHED_STALE` -- indicates that the song is cached on
      disk, but has been invalidated.
    """

    NOT_CACHED = 0
    CACHED = 1
    PERMANENTLY_CACHED = 2
    DOWNLOADING = 3
    CACHED_STALE = 4


@dataclass
class AlbumSearchQuery:
    """
    Represents a query for getting albums from an adapter. The UI will request the
    albums in pages.

    **Fields:**

    * :class:`AlbumSearchQuery.type` -- the query :class:`AlbumSearchQuery.Type`
    * :class:`AlbumSearchQuery.year_range` -- (guaranteed to only exist if ``type`` is
      :class:`AlbumSearchQuery.Type.YEAR_RANGE`) a tuple with the lower and upper bound
      (inclusive) of the album years to return
    * :class:`AlbumSearchQuery.genre` -- (guaranteed to only exist if the ``type`` is
      :class:`AlbumSearchQuery.Type.GENRE`) return albums of the given genre
    """

    class _Genre(Genre):
        def __init__(self, name: str):
            self.name = name

    class Type(Enum):
        """
        Represents a type of query. Use :class:`Adapter.supported_artist_query_types` to
        specify what search types your adapter supports.

        * :class:`AlbumSearchQuery.Type.RANDOM` -- return a random set of albums
        * :class:`AlbumSearchQuery.Type.NEWEST` -- return the most recently added albums
        * :class:`AlbumSearchQuery.Type.RECENT` -- return the most recently played
          albums
        * :class:`AlbumSearchQuery.Type.STARRED` -- return only starred albums
        * :class:`AlbumSearchQuery.Type.ALPHABETICAL_BY_NAME` -- return the albums
          sorted alphabetically by album name
        * :class:`AlbumSearchQuery.Type.ALPHABETICAL_BY_ARTIST` -- return the albums
          sorted alphabetically by artist name
        * :class:`AlbumSearchQuery.Type.YEAR_RANGE` -- return albums in the given year
          range
        * :class:`AlbumSearchQuery.Type.GENRE` -- return songs of the given genre
        """

        RANDOM = 0
        NEWEST = 1
        FREQUENT = 2
        RECENT = 3
        STARRED = 4
        ALPHABETICAL_BY_NAME = 5
        ALPHABETICAL_BY_ARTIST = 6
        YEAR_RANGE = 7
        GENRE = 8

    type: Type
    year_range: Tuple[int, int] = (2010, 2020)
    genre: Genre = _Genre("Rock")

    _strhash: Optional[str] = None

    def strhash(self) -> str:
        """
        Returns a deterministic hash of the query as a string.

        >>> AlbumSearchQuery(
        ...     AlbumSearchQuery.Type.YEAR_RANGE, year_range=(2018, 2019)
        ... ).strhash()
        '275c58cac77c5ea9ccd34ab870f59627ab98e73c'
        >>> AlbumSearchQuery(
        ...     AlbumSearchQuery.Type.YEAR_RANGE, year_range=(2018, 2020)
        ... ).strhash()
        'e5dc424e8fc3b7d9ff7878b38cbf2c9fbdc19ec2'
        >>> AlbumSearchQuery(AlbumSearchQuery.Type.STARRED).strhash()
        '861b6ff011c97d53945ca89576463d0aeb78e3d2'
        """
        if not self._strhash:
            hash_tuple: Tuple[Any, ...] = (self.type.value,)
            if self.type == AlbumSearchQuery.Type.YEAR_RANGE:
                hash_tuple += (self.year_range,)
            elif self.type == AlbumSearchQuery.Type.GENRE:
                hash_tuple += (self.genre.name,)
            self._strhash = hashlib.sha1(bytes(str(hash_tuple), "utf8")).hexdigest()
        return self._strhash


class CacheMissError(Exception):
    """
    This exception should be thrown by caching adapters when the request data is not
    available or is invalid. If some of the data is available, but not all of it, the
    ``partial_data`` parameter should be set with the partial data. If the ground truth
    adapter can't service the request, or errors for some reason, the UI will try to
    populate itself with the partial data returned in this exception (with the necessary
    error text to inform the user that retrieval from the ground truth adapter failed).
    """

    def __init__(self, *args, partial_data: Any = None):
        """
        Create a :class:`CacheMissError` exception.

        :param args: arguments to pass to the :class:`BaseException` base class.
        :param partial_data: the actual partial data for the UI to use in case of ground
            truth adapter failure.
        """
        self.partial_data = partial_data
        super().__init__(*args)


KEYRING_APP_NAME = "app.sublimemusic.SublimeMusic"


class ConfigurationStore(dict):
    """
    This defines an abstract store for all configuration parameters for a given Adapter.
    """

    MOCK = False

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._changed_secrets_store = {}

    def __repr__(self) -> str:
        values = ", ".join(f"{k}={v!r}" for k, v in sorted(self.items()))
        return f"ConfigurationStore({values})"

    def clone(self) -> "ConfigurationStore":
        configuration_store = ConfigurationStore(**copy.deepcopy(self))
        configuration_store._changed_secrets_store = copy.deepcopy(
            self._changed_secrets_store
        )
        return configuration_store

    def persist_secrets(self):
        if not keyring_imported or ConfigurationStore.MOCK:
            return

        for key, secret in self._changed_secrets_store.items():
            try:
                password_id = None
                if password_type_and_id := self.get(key):
                    if cast(List[str], password_type_and_id)[0] == "keyring":
                        password_id = password_type_and_id[1]

                if password_id is None:
                    password_id = str(uuid.uuid4())

                keyring.set_password(KEYRING_APP_NAME, password_id, secret)
                self[key] = ["keyring", password_id]
            except Exception:
                return

    def get_secret(self, key: str) -> Optional[str]:
        """
        Get the secret value in the store with the given key. This will retrieve the
        secret from whatever is configured as the underlying secret storage mechanism so
        you don't have to deal with secret storage yourself.
        """
        if secret := self._changed_secrets_store.get(key):
            return secret

        value = self.get(key)
        if not isinstance(value, list) or len(value) != 2:
            return None

        storage_type, storage_key = value
        return {
            "keyring": lambda: keyring.get_password(KEYRING_APP_NAME, storage_key),
            "plaintext": lambda: storage_key,
        }[storage_type]()

    def set_secret(self, key: str, value: str = None):
        """
        Set the secret value of the given key in the store. This should be used for
        things such as passwords or API tokens. When :class:`persist_secrets` is called,
        the secrets will be stored in whatever is configured as the underlying secret
        storage mechanism so you don't have to deal with secret storage yourself.
        """
        self._changed_secrets_store[key] = value
        self[key] = ["plaintext", value]


@dataclass
class UIInfo:
    name: str
    description: str
    icon_basename: str
    icon_dir: Optional[Path] = None

    def icon_name(self) -> str:
        return f"{self.icon_basename}-symbolic"

    def status_icon_name(self, status: str) -> str:
        return f"{self.icon_basename}-{status.lower()}-symbolic"


class Adapter(abc.ABC):
    """
    Defines the interface for a Sublime Music Adapter.

    All functions that actually retrieve data have a corresponding: ``can_``-prefixed
    property (which can be dynamic) which specifies whether or not the adapter supports
    that operation at the moment.
    """

    # Configuration and Initialization Properties
    # These functions determine how the adapter can be configured and how to
    # initialize the adapter given those configuration values.
    # ==================================================================================
    @staticmethod
    @abc.abstractmethod
    def get_ui_info() -> UIInfo:
        """
        :returns: A :class:`UIInfo` object.
        """

    @staticmethod
    @abc.abstractmethod
    def get_configuration_form(config_store: ConfigurationStore) -> Gtk.Box:
        """
        This function should return a :class:`Gtk.Box` that gets any inputs required
        from the user and uses the given ``config_store`` to store the configuration
        values.

        The ``Gtk.Box`` must expose a signal with the name ``"config-valid-changed"``
        which returns a single boolean value indicating whether or not the configuration
        is valid.

        If you don't want to implement all of the GTK logic yourself, and just want a
        simple form, then you can use the :class:`ConfigureServerForm` class to generate
        a form in a declarative manner.
        """

    @staticmethod
    @abc.abstractmethod
    def migrate_configuration(config_store: ConfigurationStore):
        """
        This function allows the adapter to migrate its configuration.
        """

    @abc.abstractmethod
    def __init__(self, config_store: ConfigurationStore, data_directory: Path):
        """
        This function should be overridden by inheritors of :class:`Adapter` and should
        be used to do whatever setup is required for the adapter.

        This should do the bare minimum to get things set up, since this blocks the main
        UI loop. If you need to do longer initialization, use the :class:`initial_sync`
        function.

        :param config: The adapter configuration. The keys of are the configuration
            parameter names as defined by the return value of the
            :class:`get_config_parameters` function. The values are the actual value of
            the configuration parameter.
        :param data_directory: the directory where the adapter can store data.  This
            directory is guaranteed to exist.
        """

    @abc.abstractmethod
    def initial_sync(self):
        """
        Perform any operations that are required to get the adapter functioning
        properly. For example, this function can be used to wait for an initial ping to
        come back from the server.
        """

    @abc.abstractmethod
    def shutdown(self):
        """
        This function is called when the app is being closed or the server is changing.
        This should be used to clean up anything that is necessary such as writing a
        cache to disk, disconnecting from a server, etc.
        """

    # Usage Properties
    # These properties determine how the adapter can be used and how quickly
    # data can be expected from this adapter.
    # ==================================================================================
    @property
    def can_be_cached(self) -> bool:
        """
        Whether or not this adapter can be used as the ground-truth adapter behind a
        caching adapter.

        The default is ``True``, since most adapters will want to take advantage of the
        built-in filesystem cache.
        """
        return True

    @property
    @staticmethod
    def can_be_ground_truth() -> bool:
        """
        Whether or not this adapter can be used as a ground truth adapter.
        """
        return True

    # Network Properties
    # These properties determine whether or not the adapter requires connection over a
    # network and whether the underlying server can be pinged.
    # ==================================================================================
    @property
    def is_networked(self) -> bool:
        """
        Whether or not this adapter operates over the network. This will be used to
        determine whether or not some of the offline/online management features should
        be enabled.
        """
        return True

    def on_offline_mode_change(self, offline_mode: bool):
        """
        This function should be used to handle any operations that need to be performed
        when Sublime Music goes from online to offline mode or vice versa.
        """

    @property
    @abc.abstractmethod
    def ping_status(self) -> bool:
        """
        If the adapter :class:`is_networked`, then this function should return whether
        or not the server can be pinged. This function must provide an answer
        *instantly* (it can't actually ping the server). This function is called *very*
        often, and even a few milliseconds delay stacks up quickly and can block the UI
        thread.

        One option is to ping the server every few seconds and cache the result of the
        ping and use that as the result of this function.
        """

    # Availability Properties
    # These properties determine if what things the adapter can be used to do. These
    # properties can be dynamic based on things such as underlying API version, or other
    # factors like that. However, these properties should not be dependent on the
    # connection state. After the initial sync, these properties are expected to be
    # constant.
    # ==================================================================================
    # Playlists
    @property
    def can_get_playlists(self) -> bool:
        """
        Whether or not the adapter supports :class:`get_playlist`.
        """
        return False

    @property
    def can_get_playlist_details(self) -> bool:
        """
        Whether or not the adapter supports :class:`get_playlist_details`.
        """
        return False

    @property
    def can_create_playlist(self) -> bool:
        """
        Whether or not the adapter supports :class:`create_playlist`.
        """
        return False

    @property
    def can_update_playlist(self) -> bool:
        """
        Whether or not the adapter supports :class:`update_playlist`.
        """
        return False

    @property
    def can_delete_playlist(self) -> bool:
        """
        Whether or not the adapter supports :class:`delete_playlist`.
        """
        return False

    # Downloading/streaming cover art and songs
    @property
    def supported_schemes(self) -> Iterable[str]:
        """
        Specifies a collection of scheme names that can be provided by the adapter for a
        given resource (song or cover art) right now.

        Examples of values that could be provided include ``http``, ``https``, ``file``,
        or ``ftp``.
        """
        return ()

    @property
    def can_get_cover_art_uri(self) -> bool:
        """
        Whether or not the adapter supports :class:`get_cover_art_uri`.
        """
        return False

    @property
    def can_get_song_file_uri(self) -> bool:
        """
        Whether or not the adapter supports :class:`get_song_file_uri`.
        """
        return False

    @property
    def can_get_song_stream_uri(self) -> bool:
        """
        Whether or not the adapter supports :class:`get_song_stream_uri`.
        """
        return False

    # Songs
    @property
    def can_get_song_details(self) -> bool:
        """
        Whether or not the adapter supports :class:`get_song_details`.
        """
        return False

    @property
    def can_scrobble_song(self) -> bool:
        """
        Whether or not the adapter supports :class:`scrobble_song`.
        """
        return False

    # Artists
    @property
    def supported_artist_query_types(self) -> Set[AlbumSearchQuery.Type]:
        """
        A set of the query types that this adapter can service.

        :returns: A set of :class:`AlbumSearchQuery.Type` objects.
        """
        return set()

    @property
    def can_get_artists(self) -> bool:
        """
        Whether or not the adapter supports :class:`get_aritsts`.
        """
        return False

    @property
    def can_get_artist(self) -> bool:
        """
        Whether or not the adapter supports :class:`get_aritst`.
        """
        return False

    @property
    def can_get_ignored_articles(self) -> bool:
        """
        Whether or not the adapter supports :class:`get_ignored_articles`.
        """
        return False

    # Albums
    @property
    def can_get_albums(self) -> bool:
        """
        Whether or not the adapter supports :class:`get_albums`.
        """
        return False

    @property
    def can_get_album(self) -> bool:
        """
        Whether or not the adapter supports :class:`get_album`.
        """
        return False

    # Browse directories
    @property
    def can_get_directory(self) -> bool:
        """
        Whether or not the adapter supports :class:`get_directory`.
        """
        return False

    # Genres
    @property
    def can_get_genres(self) -> bool:
        """
        Whether or not the adapter supports :class:`get_genres`.
        """
        return False

    # Play Queue
    @property
    def can_get_play_queue(self) -> bool:
        """
        Whether or not the adapter supports :class:`get_play_queue`.
        """
        return False

    @property
    def can_save_play_queue(self) -> bool:
        """
        Whether or not the adapter supports :class:`save_play_queue`.
        """
        return False

    # Search
    @property
    def can_search(self) -> bool:
        """
        Whether or not the adapter supports :class:`search`.
        """
        return False

    # Data Retrieval Methods
    # These properties determine if what things the adapter can be used to do
    # at the current moment.
    # ==================================================================================
    def get_playlists(self) -> Sequence[Playlist]:
        """
        Get a list of all of the playlists known by the adapter.

        :returns: A list of all of the
            :class:`sublime_music.adapter.api_objects.Playlist` objects known to the
            adapter.
        """
        raise self._check_can_error("get_playlists")

    def get_playlist_details(self, playlist_id: str) -> Playlist:
        """
        Get the details for the given ``playlist_id``. If the playlist_id does not
        exist, then this function should throw an exception.

        :param playlist_id: The ID of the playlist to retrieve.
        :returns: A :class:`sublime_music.adapter.api_objects.Play` object for the given
            playlist.
        """
        raise self._check_can_error("get_playlist_details")

    def create_playlist(
        self, name: str, songs: Sequence[Song] = None
    ) -> Optional[Playlist]:
        """
        Creates a playlist of the given name with the given songs.

        :param name: The human-readable name of the playlist.
        :param songs: A list of songs that should be included in the playlist.
        :returns: A :class:`sublime_music.adapter.api_objects.Playlist` object for the
            created playlist. If getting this information will incurr network overhead,
            then just return ``None``.
        """
        raise self._check_can_error("create_playlist")

    def update_playlist(
        self,
        playlist_id: str,
        name: str = None,
        comment: str = None,
        public: bool = None,
        song_ids: Sequence[str] = None,
    ) -> Playlist:
        """
        Updates a given playlist. If a parameter is ``None``, then it will be ignored
        and no updates will occur to that field.

        :param playlist_id: The human-readable name of the playlist.
        :param name: The human-readable name of the playlist.
        :param comment: The playlist comment.
        :param public: This is very dependent on the adapter, but if the adapter has a
            shared/public vs. not shared/private playlists concept, setting this to
            ``True`` will make the playlist shared/public.
        :param song_ids: A list of song IDs that should be included in the playlist.
        :returns: A :class:`sublime_music.adapter.api_objects.Playlist` object for the
            updated playlist.
        """
        raise self._check_can_error("update_playlist")

    def delete_playlist(self, playlist_id: str):
        """
        Deletes the given playlist.

        :param playlist_id: The human-readable name of the playlist.
        """
        raise self._check_can_error("delete_playlist")

    def get_cover_art_uri(self, cover_art_id: str, scheme: str, size: int) -> str:
        """
        Get a URI for a given ``cover_art_id``.

        :param cover_art_id: The song, album, or artist ID.
        :param scheme: The URI scheme that should be returned. It is guaranteed that
            ``scheme`` will be one of the schemes returned by
            :class:`supported_schemes`.
        :param size: The size of image to return. Denotes the max width or max height
            (whichever is larger).
        :returns: The URI as a string.
        """
        raise self._check_can_error("get_cover_art_uri")

    def get_song_file_uri(self, song_id: str, schemes: Iterable[str]) -> str:
        """
        Get a URI for a given song. This URI must give the full file.

        :param song_id: The ID of the song to get a URI for.
        :param schemes: A set of URI schemes that can be returned. It is guaranteed that
            all of the items in ``schemes`` will be one of the schemes returned by
            :class:`supported_schemes`.
        :returns: The URI for the given song.
        """
        raise self._check_can_error("get_song_file_uri")

    def get_song_stream_uri(self, song_id: str) -> str:
        """
        Get a URI for streaming the given song.

        :param song_id: The ID of the song to get the stream URI for.
        :returns: the stream URI for the given song.
        """
        raise self._check_can_error("get_song_stream_uri")

    def get_song_details(self, song_id: str) -> Song:
        """
        Get the details for a given song ID.

        :param song_id: The ID of the song to get the details for.
        :returns: The :class:`sublime_music.adapters.api_objects.Song`.
        """
        raise self._check_can_error("get_song_details")

    def scrobble_song(self, song: Song):
        """
        Scrobble the given song.

        :params song: The :class:`sublime_music.adapters.api_objects.Song` to scrobble.
        """
        raise self._check_can_error("scrobble_song")

    def get_artists(self) -> Sequence[Artist]:
        """
        Get a list of all of the artists known to the adapter.

        :returns: A list of all of the :class:`sublime_music.adapter.api_objects.Artist`
            objects known to the adapter.
        """
        raise self._check_can_error("get_artists")

    def get_artist(self, artist_id: str) -> Artist:
        """
        Get the details for the given artist ID.

        :param artist_id: The ID of the artist to get the details for.
        :returns: The :classs`sublime_music.adapters.api_objects.Artist`
        """
        raise self._check_can_error("get_artist")

    def get_ignored_articles(self) -> Set[str]:
        """
        Get the list of articles to ignore when sorting artists by name.

        :returns: A set of articles (i.e. The, A, El, La, Los) to ignore when sorting
            artists.
        """
        raise self._check_can_error("get_ignored_articles")

    def get_albums(
        self, query: AlbumSearchQuery, sort_direction: str = "ascending"
    ) -> Sequence[Album]:
        """
        Get a list of all of the albums known to the adapter for the given query.

        .. note::

           This request is not paged. You should do any page management to get all of
           the albums matching the query internally.

        :param query: An :class:`AlbumSearchQuery` object representing the types of
            albums to return.
        :returns: A list of all of the :class:`sublime_music.adapter.api_objects.Album`
            objects known to the adapter that match the query.
        """
        raise self._check_can_error("get_albums")

    def get_album(self, album_id: str) -> Album:
        """
        Get the details for the given album ID.

        :param album_id: The ID of the album to get the details for.
        :returns: The :classs`sublime_music.adapters.api_objects.Album`
        """
        raise self._check_can_error("get_album")

    def get_directory(self, directory_id: str) -> Directory:
        """
        Return a Directory object representing the song files and directories in the
        given directory. This may not make sense for your adapter (for example, if
        there's no actual underlying filesystem). In that case, make sure to set
        :class:`can_get_directory` to ``False``.

        :param directory_id: The directory to retrieve. If the special value ``"root"``
            is given, the adapter should list all of the directories at the root of the
            filesystem tree.
        :returns: A list of the :class:`sublime_music.adapter.api_objects.Directory` and
            :class:`sublime_music.adapter.api_objects.Song` objects in the given
            directory.
        """
        raise self._check_can_error("get_directory")

    def get_genres(self) -> Sequence[Genre]:
        """
        Get a list of the genres known to the adapter.

        :returns: A list of all of the :classs`sublime_music.adapter.api_objects.Genre`
            objects known to the adapter.
        """
        raise self._check_can_error("get_genres")

    def get_play_queue(self) -> Optional[PlayQueue]:
        """
        Returns the state of the play queue for this user. This could be used to restore
        the play queue from the cloud.

        :returns: The cloud-saved play queue as a
            :class:`sublime_music.adapter.api_objects.PlayQueue` object.
        """
        raise self._check_can_error("get_play_queue")

    def save_play_queue(
        self,
        song_ids: Sequence[str],
        current_song_index: int = None,
        position: timedelta = None,
    ):
        """
        Save the current play queue to the cloud.

        :param song_ids: A list of the song IDs in the queue.
        :param current_song_index: The index of the song that is currently being played.
        :param position: The current position in the song.
        """
        raise self._check_can_error("can_save_play_queue")

    def search(self, query: str) -> SearchResult:
        """
        Return search results fro the given query.

        :param query: The query string.
        :returns: A :class:`sublime_music.adapters.api_objects.SearchResult` object
            representing the results of the search.
        """
        raise self._check_can_error("search")

    @staticmethod
    def _check_can_error(method_name: str) -> NotImplementedError:
        return NotImplementedError(
            f"Adapter.{method_name} called. "
            "Did you forget to check that can_{method_name} is True?"
        )


class CachingAdapter(Adapter):
    """
    Defines an adapter that can be used as a cache for another adapter.

    A caching adapter sits "in front" of a non-caching adapter and the UI will attempt
    to retrieve the data from the caching adapter before retrieving it from the
    non-caching adapter. (The exception is when the UI requests that the data come
    directly from the ground truth adapter, in which case the cache will be bypassed.)

    Caching adapters *must* be able to service requests instantly, or nearly instantly
    (in most cases, this means that the data must come directly from the local
    filesystem).
    """

    @abc.abstractmethod
    def __init__(self, config: dict, data_directory: Path, is_cache: bool = False):
        """
        This function should be overridden by inheritors of :class:`CachingAdapter` and
        should be used to do whatever setup is required for the adapter.

        :param config: The adapter configuration. The keys of are the configuration
            parameter names as defined by the return value of the
            :class:`get_config_parameters` function. The values are the actual value of
            the configuration parameter.
        :param data_directory: the directory where the adapter can store data.  This
            directory is guaranteed to exist.
        :param is_cache: whether or not the adapter is being used as a cache.
        """

    ping_status = True

    # Data Ingestion Methods
    # ==================================================================================
    class CachedDataKey(Enum):
        ALBUM = "album"
        ALBUMS = "albums"
        ARTIST = "artist"
        ARTISTS = "artists"
        COVER_ART_FILE = "cover_art_file"
        DIRECTORY = "directory"
        GENRE = "genre"
        GENRES = "genres"
        IGNORED_ARTICLES = "ignored_articles"
        PLAYLIST_DETAILS = "get_playlist_details"
        PLAYLISTS = "get_playlists"
        SEARCH_RESULTS = "search_results"
        SONG = "song"
        SONG_FILE = "song_file"
        SONG_FILE_PERMANENT = "song_file_permanent"

        # These are only for clearing the cache, and will only do deletion
        ALL_SONGS = "all_songs"
        EVERYTHING = "everything"

    @abc.abstractmethod
    def ingest_new_data(self, data_key: CachedDataKey, param: Optional[str], data: Any):
        """
        This function will be called after the fallback, ground-truth adapter returns
        new data. This normally will happen if this adapter has a cache miss or if the
        UI forces retrieval from the ground-truth adapter.

        :param data_key: the type of data to be ingested.
        :param param: a string that uniquely identify the data to be ingested. For
            example, with playlist details, this will be the playlist ID. If that
            playlist ID is requested again, the adapter should service that request, but
            it should not service a request for a different playlist ID.

            For the playlist list, this will be none since there are no parameters to
            that request.
        :param data: the data that was returned by the ground truth adapter.
        """

    @abc.abstractmethod
    def invalidate_data(self, data_key: CachedDataKey, param: Optional[str]):
        """
        This function will be called if the adapter should invalidate some of its data.
        This should not destroy the invalidated data. If invalid data is requested, a
        ``CacheMissError`` should be thrown, but the old data should be included in the
        ``partial_data`` field of the error.

        :param data_key: the type of data to be invalidated.
        :param params: the parameters that uniquely identify the data to be invalidated.
            For example, with playlist details, this will be the playlist ID.

            For the playlist list, this will be none since there are no parameters to
            that request.
        """

    @abc.abstractmethod
    def delete_data(self, data_key: CachedDataKey, param: Optional[str]):
        """
        This function will be called if the adapter should delete some of its data.
        This should destroy the data. If the deleted data is requested, a
        ``CacheMissError`` should be thrown with no data in the ``partial_data`` field.

        :param data_key: the type of data to be deleted.
        :param params: the parameters that uniquely identify the data to be invalidated.
            For example, with playlist details, this will be the playlist ID.

            For the playlist list, this will be none since there are no parameters to
            that request.
        """

    # Cache-Specific Methods
    # ==================================================================================
    @abc.abstractmethod
    def get_cached_statuses(
        self, song_ids: Sequence[str]
    ) -> Dict[str, SongCacheStatus]:
        """
        Returns the cache statuses for the given list of songs. See the
        :class:`SongCacheStatus` documentation for more details about what each status
        means.

        :params songs: The songs to get the cache status for.
        :returns: A dictionary of song ID to :class:`SongCacheStatus` objects for each
            of the songs.
        """



import base64
import io
import logging
import mimetypes
import multiprocessing
import os
import socket
from datetime import timedelta
from typing import Any, Callable, cast, Dict, Optional, Set, Tuple, Type, Union
from urllib.parse import urlparse
from uuid import UUID

from gi.repository import GLib

from .base import Player, PlayerDeviceEvent, PlayerEvent
from ..adapters import AdapterManager
from ..adapters.api_objects import Song

try:
    import pychromecast

    chromecast_imported = True
except Exception:
    chromecast_imported = False

try:
    import bottle

    bottle_imported = True
except Exception:
    bottle_imported = False

SERVE_FILES_KEY = "Serve Local Files to Chromecasts on the LAN"
LAN_PORT_KEY = "LAN Server Port Number"


class ChromecastPlayer(Player):
    name = "Chromecast"
    can_start_playing_with_no_latency = False

    @property
    def enabled(self) -> bool:
        return chromecast_imported

    @staticmethod
    def get_configuration_options() -> Dict[str, Union[Type, Tuple[str, ...]]]:
        if not bottle_imported:
            return {}
        return {SERVE_FILES_KEY: bool, LAN_PORT_KEY: int}

    @property
    def supported_schemes(self) -> Set[str]:
        schemes = {"http", "https"}
        if bottle_imported and self.config.get(SERVE_FILES_KEY):
            schemes.add("file")
        return schemes

    _timepos = 0.0

    def __init__(
        self,
        on_timepos_change: Callable[[Optional[float]], None],
        on_track_end: Callable[[], None],
        on_player_event: Callable[[PlayerEvent], None],
        player_device_change_callback: Callable[[PlayerDeviceEvent], None],
        config: Dict[str, Union[str, int, bool]],
    ):
        self.server_process: Optional[multiprocessing.Process] = None
        self.on_timepos_change = on_timepos_change
        self.on_track_end = on_track_end
        self.on_player_event = on_player_event
        self.player_device_change_callback = player_device_change_callback

        self.change_settings(config)

        if chromecast_imported:
            self._chromecasts: Dict[UUID, pychromecast.Chromecast] = {}
            self._current_chromecast: Optional[pychromecast.Chromecast] = None

            self.stop_get_chromecasts = None
            self.refresh_players()

    def chromecast_discovered_callback(self, chromecast: Any):
        chromecast = cast(pychromecast.Chromecast, chromecast)
        self._chromecasts[chromecast.device.uuid] = chromecast
        self.player_device_change_callback(
            PlayerDeviceEvent(
                PlayerDeviceEvent.Delta.ADD,
                type(self),
                str(chromecast.device.uuid),
                chromecast.device.friendly_name,
            )
        )

    def change_settings(self, config: Dict[str, Union[str, int, bool]]):
        if not chromecast_imported:
            return

        self.config = config
        if bottle_imported and self.config.get(SERVE_FILES_KEY):
            # Try and terminate the existing process if it exists.
            if self.server_process is not None:
                try:
                    self.server_process.terminate()
                except Exception:
                    pass

            self.server_process = multiprocessing.Process(
                target=self._run_server_process,
                args=("0.0.0.0", self.config.get(LAN_PORT_KEY)),
            )
            self.server_process.start()

    def refresh_players(self):
        if not chromecast_imported:
            return

        if self.stop_get_chromecasts is not None:
            self.stop_get_chromecasts.cancel()

        for id_, chromecast in self._chromecasts.items():
            self.player_device_change_callback(
                PlayerDeviceEvent(
                    PlayerDeviceEvent.Delta.REMOVE,
                    type(self),
                    str(id_),
                    chromecast.device.friendly_name,
                )
            )

        self._chromecasts = {}

        self.stop_get_chromecasts = pychromecast.get_chromecasts(
            blocking=False, callback=self.chromecast_discovered_callback
        )

    def set_current_device_id(self, device_id: str):
        self._current_chromecast = self._chromecasts[UUID(device_id)]
        self._current_chromecast.media_controller.register_status_listener(self)
        self._current_chromecast.register_status_listener(self)
        self._current_chromecast.wait()

    def new_cast_status(self, status: Any):
        assert self._current_chromecast
        self.on_player_event(
            PlayerEvent(
                PlayerEvent.EventType.VOLUME_CHANGE,
                str(self._current_chromecast.device.uuid),
                volume=(status.volume_level * 100 if not status.volume_muted else 0),
            )
        )

        # This normally happens when "Stop Casting" is pressed in the Google
        # Home app.
        if status.session_id is None:
            self.on_player_event(
                PlayerEvent(
                    PlayerEvent.EventType.PLAY_STATE_CHANGE,
                    str(self._current_chromecast.device.uuid),
                    playing=False,
                )
            )
            self.on_player_event(
                PlayerEvent(
                    PlayerEvent.EventType.DISCONNECT,
                    str(self._current_chromecast.device.uuid),
                )
            )
            self.song_loaded = False

    time_increment_order_token = 0

    def new_media_status(self, status: Any):
        # Detect the end of a track and go to the next one.
        if (
            status.idle_reason == "FINISHED"
            and status.player_state == "IDLE"
            and self._timepos > 0
        ):
            self.on_track_end()
            return

        self.song_loaded = True

        self._timepos = status.current_time

        assert self._current_chromecast
        self.on_player_event(
            PlayerEvent(
                PlayerEvent.EventType.PLAY_STATE_CHANGE,
                str(self._current_chromecast.device.uuid),
                playing=(status.player_state in ("PLAYING", "BUFFERING")),
            )
        )

        def increment_time(order_token: int):
            if self.time_increment_order_token != order_token or not self.playing:
                return

            self._timepos += 0.5
            self.on_timepos_change(self._timepos)
            GLib.timeout_add(500, increment_time, order_token)

        self.time_increment_order_token += 1
        GLib.timeout_add(500, increment_time, self.time_increment_order_token)

    def shutdown(self):
        if self.server_process:
            self.server_process.terminate()

        try:
            self._current_chromecast.disconnect()
        except Exception:
            pass

    _serving_song_id = multiprocessing.Array("c", 1024)  # huge buffer, just in case
    _serving_token = multiprocessing.Array("c", 16)

    def _run_server_process(self, host: str, port: int):
        app = bottle.Bottle()

        @app.route("/")
        def index() -> str:
            return """
            <h1>Sublime Music Local Music Server</h1>
            <p>
                Sublime Music uses this port as a server for serving music Chromecasts
                on the same LAN.
            </p>
            """

        @app.route("/s/<token>")
        def stream_song(token: str) -> bytes:
            if token != self._serving_token.value.decode():
                raise bottle.HTTPError(status=401, body="Invalid token.")

            song = AdapterManager.get_song_details(
                self._serving_song_id.value.decode()
            ).result()
            filename = AdapterManager.get_song_file_uri(song)
            with open(filename[7:], "rb") as fin:
                song_buffer = io.BytesIO(fin.read())

            content_type = mimetypes.guess_type(filename)[0]
            bottle.response.set_header("Content-Type", content_type)
            bottle.response.set_header("Accept-Ranges", "bytes")
            return song_buffer.read()

        bottle.run(app, host=host, port=port)

    @property
    def playing(self) -> bool:
        if (
            not self._current_chromecast
            or not self._current_chromecast.media_controller
        ):
            return False
        return self._current_chromecast.media_controller.status.player_is_playing

    def get_volume(self) -> float:
        if self._current_chromecast:
            # The volume is in the range [0, 1]. Multiply by 100 to get to [0, 100].
            return self._current_chromecast.status.volume_level * 100
        else:
            return 100

    def set_volume(self, volume: float):
        if self._current_chromecast:
            # volume value is in [0, 100]. Convert to [0, 1] for Chromecast.
            self._current_chromecast.set_volume(volume / 100)

    def get_is_muted(self) -> bool:
        if not self._current_chromecast:
            return False
        return self._current_chromecast.volume_muted

    def set_muted(self, muted: bool):
        if not self._current_chromecast:
            return
        self._current_chromecast.set_volume_muted(muted)

    def play_media(self, uri: str, progress: timedelta, song: Song):
        assert self._current_chromecast
        scheme = urlparse(uri).scheme
        if scheme == "file":
            token = base64.b16encode(os.urandom(8))
            self._serving_token.value = token
            self._serving_song_id.value = song.id.encode()

            # If this fails, then we are basically screwed, so don't care if it blows
            # up.
            # TODO (#129): this does not work properly when on VPNs when the DNS is
            # piped over the VPN tunnel.
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            s.connect(("8.8.8.8", 80))
            host_ip = s.getsockname()[0]
            s.close()

            uri = f"http://{host_ip}:{self.config.get(LAN_PORT_KEY)}/s/{token.decode()}"
            logging.info("Serving {song.name} at {uri}")

        assert AdapterManager._instance
        networked_scheme_priority = ("https", "http")
        scheme = sorted(
            AdapterManager._instance.ground_truth_adapter.supported_schemes,
            key=lambda s: networked_scheme_priority.index(s),
        )[0]
        cover_art_url = AdapterManager.get_cover_art_uri(
            song.cover_art, scheme, size=1000
        ).result()
        self._current_chromecast.media_controller.play_media(
            uri,
            # Just pretend that whatever we send it is mp3, even if it isn't.
            "audio/mp3",
            current_time=progress.total_seconds(),
            title=song.title,
            thumb=cover_art_url,
            metadata={
                "metadataType": 3,
                "albumName": song.album.name if song.album else None,
                "artist": song.artist.name if song.artist else None,
                "trackNumber": song.track,
            },
        )

        # Make sure to clear out the cache duration state.
        self.on_player_event(
            PlayerEvent(
                PlayerEvent.EventType.STREAM_CACHE_PROGRESS_CHANGE,
                str(self._current_chromecast.device.uuid),
                stream_cache_duration=0,
            )
        )
        self._timepos = progress.total_seconds()

    def pause(self):
        if self._current_chromecast and self._current_chromecast.media_controller:
            self._current_chromecast.media_controller.pause()

    def play(self):
        if self._current_chromecast and self._current_chromecast.media_controller:
            self._current_chromecast.media_controller.play()

    def seek(self, position: timedelta):
        if not self._current_chromecast:
            return

        do_pause = not self.playing
        self._current_chromecast.media_controller.seek(position.total_seconds())
        if do_pause:
            self.pause()

    def _wait_for_playing(self):
        pass



from pathlib import Path

from setuptools import find_packages, setup

here = Path(__file__).parent.resolve()

with open(here.joinpath("README.rst"), encoding="utf-8") as f:
    long_description = f.read()

# Find the version
with open(here.joinpath("sublime_music", "__init__.py")) as f:
    for line in f:
        if line.startswith("__version__"):
            version = eval(line.split()[-1])
            break

package_data_dirs = [
    here.joinpath("sublime_music", "adapters", "icons"),
    here.joinpath("sublime_music", "adapters", "images"),
    here.joinpath("sublime_music", "adapters", "subsonic", "icons"),
    here.joinpath("sublime_music", "dbus", "mpris_specs"),
    here.joinpath("sublime_music", "ui", "icons"),
    here.joinpath("sublime_music", "ui", "images"),
]
package_data_files = []
for data_dir in package_data_dirs:
    for file in data_dir.iterdir():
        package_data_files.append(str(file))

setup(
    name="sublime_music",
    version=version,
    url="https://gitlab.com/sublime-music/sublime-music",
    description="A native GTK *sonic client.",
    long_description=long_description,
    author="Sumner Evans",
    author_email="inquiries@sumnerevans.com",
    license="GPL3",
    classifiers=[
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        "Development Status :: 3 - Alpha",
        # Indicate who your project is intended for
        "Intended Audience :: End Users/Desktop",
        "Operating System :: POSIX",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        "Programming Language :: Python :: 3.8",
    ],
    keywords="airsonic subsonic libresonic gonic music",
    packages=find_packages(exclude=["tests"]),
    package_data={"sublime_music": ["ui/app_styles.css", *package_data_files]},
    install_requires=[
        "bleach",
        "dataclasses-json",
        "deepdiff",
        "fuzzywuzzy",
        'osxmmkeys ; sys_platform=="darwin"',
        "peewee",
        "PyGObject",
        "python-dateutil",
        "python-Levenshtein",
        "python-mpv",
        "requests",
        "semver",
    ],
    extras_require={
        "keyring": ["keyring"],
        "chromecast": ["pychromecast"],
        "server": ["bottle"],
    },
    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and
    # allow pip to create the appropriate form of executable for the target
    # platform.
    entry_points={"console_scripts": ["sublime-music=sublime_music.__main__:main"]},
)



from gi.repository import Gtk


class LoadError(Gtk.Box):
    def __init__(
        self, entity_name: str, action: str, has_data: bool, offline_mode: bool
    ):
        Gtk.Box.__init__(self, orientation=Gtk.Orientation.VERTICAL)

        self.pack_start(Gtk.Box(), True, True, 0)

        inner_box = Gtk.Box(
            orientation=Gtk.Orientation.HORIZONTAL, name="load-error-box"
        )

        inner_box.pack_start(Gtk.Box(), True, True, 0)

        error_and_button_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        icon_and_button_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)

        if offline_mode:
            icon_name = "cloud-offline-symbolic"
            label = f"{entity_name} may be incomplete.\n" if has_data else ""
            label += f"Go online to {action}."
        else:
            icon_name = "network-error-symbolic"
            label = f"Error attempting to {action}."

        self.image = Gtk.Image.new_from_icon_name(icon_name, Gtk.IconSize.DIALOG)
        self.image.set_name("load-error-image")
        icon_and_button_box.add(self.image)

        self.label = Gtk.Label(label=label, name="load-error-label")
        icon_and_button_box.add(self.label)

        error_and_button_box.add(icon_and_button_box)

        button_centerer_box = Gtk.Box()
        button_centerer_box.pack_start(Gtk.Box(), True, True, 0)

        if offline_mode:
            go_online_button = Gtk.Button(label="Go Online")
            go_online_button.set_action_name("app.go-online")
            button_centerer_box.add(go_online_button)
        else:
            retry_button = Gtk.Button(label="Retry")
            retry_button.set_action_name("app.refresh-window")
            button_centerer_box.add(retry_button)

        button_centerer_box.pack_start(Gtk.Box(), True, True, 0)
        error_and_button_box.add(button_centerer_box)

        inner_box.add(error_and_button_box)

        inner_box.pack_start(Gtk.Box(), True, True, 0)

        self.add(inner_box)

        self.pack_start(Gtk.Box(), True, True, 0)
